/*
  QtCurve (C) Craig Drummond, 2003 Craig.Drummond@lycos.co.uk

  Based upon FreeCurve/B???Curve
  ----

 * B???curve theme engine
 * Copyright (C) 2001 R?d H?t, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Written by Owen Taylor <otaylor@r?dh?t.com>
 * modified by Alexander Larsson <alexl@r?dh?t.com>
 */

#include <gmodule.h>
#include <gtk/gtk.h>
#include <gtk/gtkstyle.h>
#include <gdk/gdktypes.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "qtcurve.h"
#include "qt_settings.h"
#include "common.h"
#include "config.h"

#if GTK_MAJOR_VERSION>1
#define QT_CURVE_STYLE_VAR QtCurveStyle *qtcurve_style = QTCURVE_STYLE(style);
#define QT_STYLE style
#define PCONST const
#define WIDGET_TYPE_NAME(xx) (widget && !strcmp(g_type_name (G_TYPE_FROM_INSTANCE(widget)), (xx)))
#define FN_CHECK g_return_if_fail(GTK_IS_STYLE(style)); g_return_if_fail(window != NULL);
#elif GTK_MAJOR_VERSION==1
#define QT_CURVE_STYLE_VAR QtCurveStyle *qtcurve_style = style->engine_data;
#define QT_STYLE style->klass
#define PCONST
#define WIDGET_TYPE_NAME(xx) (FALSE)
#define FN_CHECK g_return_if_fail(window != NULL);
#endif

#define SCALE_SIZE 5
#define LARGE_ARR_WIDTH  7
#define LARGE_ARR_HEIGHT (global.v_arrow ? 5 : 4)
#define SMALL_ARR_WIDTH  5
#define SMALL_ARR_HEIGHT (global.v_arrow ? 4 : 3)

#if 0
static void display_widget(GtkWidget *widget, int level)
{
    if(level>=0)
    {
        printf("%s(%s) ", widget ? gtk_type_name(GTK_WIDGET_TYPE(widget)) : "NULL",
               widget && widget->name ? widget->name : "NULL");
        if(widget && widget->parent)
            display_widget(widget->parent, --level);
        else
            printf("\n");
    }
    else
        printf("\n");
}
#endif

static GtkStyleClass *parent_class;
static QtCurveGlobalData global;

#if GTK_MAJOR_VERSION==1
typedef struct _GtkBorder      GtkBorder;
struct _GtkBorder
{
    gint left;
    gint right;
    gint top;
    gint bottom;
};
#endif

static GtkRequisition default_option_indicator_size    = { 6, 13 };
static GtkBorder      default_option_indicator_spacing = { 7, 5, 1, 1 };

#define DETAIL(xx) ((detail) &&(!strcmp(xx, detail)))

#define QTC_PANED "paned-qtc"
#define IS_QTC_PANED DETAIL(QTC_PANED)

static void shade(GdkColor *a, GdkColor *b, float k);

static gboolean use_button_color(const gchar *detail)
{
    return detail &&( 0==strcmp(detail, "optionmenu") ||
                       0==strcmp(detail, "button") ||
                       ((global.border_button || global.rounded) && (0==strcmp(detail, "buttondefault") ||
                                                                     0==strcmp(detail, "togglebuttondefault"))) ||
                       0==strcmp(detail, "togglebutton") ||
                       0==strcmp(detail, "hscale") ||
                       0==strcmp(detail, "vscale") ||
                       0==strcmp(detail, "spinbutton") ||
                       0==strcmp(detail, "spinbutton_up") ||
                       0==strcmp(detail, "spinbutton_down") ||
                       0==strcmp(detail, "slider") ||
                       0==strcmp(detail, "vscrollbar") ||
                       0==strcmp(detail, "hscrollbar") ||
                       0==strcmp(detail, QTC_PANED) );
}

#define QT_COL_EQ(A, B)(abs(A-B)<(3<<8))

#define QT_CUSTOM_COLOR_BUTTON(style) \
    (style && \
    !(QT_COL_EQ(qt_settings.colors[COLOR_WINDOW].red,(style->bg[GTK_STATE_NORMAL].red)) && \
      QT_COL_EQ(qt_settings.colors[COLOR_WINDOW].green,(style->bg[GTK_STATE_NORMAL].green)) && \
      QT_COL_EQ(qt_settings.colors[COLOR_WINDOW].blue,(style->bg[GTK_STATE_NORMAL].blue))))

static void shade_colors(GdkColor *base, GdkColor *vals)
{
    QTC_SHADES

    int i;

    for(i=0; i<6; ++i)
        shade(base, &vals[i], QTC_SHADE(global.contrast, i));
    vals[6]=*base;
}

static GdkGC * realize_color(GtkStyle *style, GdkColor *color);

#define GEN_GCS(style, colors, gcs) \
{ \
    int loop_count; \
 \
    for(loop_count=0;loop_count<7;++loop_count) \
        gcs[loop_count] = realize_color(style, &colors[loop_count]); \
}

#define RELEASE_GCS(gcs) \
{ \
    int loop_count; \
 \
    for(loop_count=0; loop_count<7;++loop_count) \
        gtk_gc_release(gcs[loop_count]); \
}

static GdkGCValues qtc_mid_color_old_values;
static GdkGC       *qtc_mid_color_gc=NULL;

#if GTK_MAJOR_VERSION==1
void gdk_rgb_find_color(GdkColormap *colormap, GdkColor *color)
{
    color->pixel = gdk_rgb_xpixel_from_rgb(((color->red >> 8) << 16) | ((color->green >> 8) << 8) | (color->blue >> 8));
}
#endif

inline int limit(double c)
{
    return c < 0.0
               ? 0
               : c > 65536.0
                     ? 65536
                    : (int)c;
}

#define QTC_SET_MID_COLOR_FACTOR(GC, A, B, factor) \
{ \
    GdkColor mid_color=B; \
 \
    qtc_mid_color_gc=GC;\
    mid_color.red=(A.red+limit(B.red*factor))>>1; \
    mid_color.green=(A.green+limit(B.green*factor))>>1; \
    mid_color.blue=(A.blue+limit(B.blue*factor))>>1; \
    gdk_gc_get_values(qtc_mid_color_gc, &qtc_mid_color_old_values); \
    gdk_rgb_find_color(style->colormap, &mid_color); \
    gdk_gc_set_foreground(qtc_mid_color_gc, &mid_color); \
    if(area) \
        gdk_gc_set_clip_rectangle(qtc_mid_color_gc, area); \
}

#define QTC_SET_MID_COLOR(GC, A, B) QTC_SET_MID_COLOR_FACTOR(GC, A, B, 1.0)

#define QTC_UNSET_MID_COLOR \
    gdk_gc_set_foreground(qtc_mid_color_gc, &qtc_mid_color_old_values.foreground); \
    if(area) \
        gdk_gc_set_clip_rectangle(qtc_mid_color_gc, NULL);

#define QTC_ARROW_STATE(state) (GTK_STATE_ACTIVE==state ? GTK_STATE_NORMAL : state)

typedef enum
{
    ROUNDED_NONE,
    ROUNDED_TOP,
    ROUNDED_BOTTOM,
    ROUNDED_LEFT,
    ROUNDED_RIGHT,
    ROUNDED_ALL
} ERound;

static gboolean is_button_on_toolbar(GtkWidget *widget)
{
    return widget && widget->parent && GTK_IS_BUTTON(widget) && GTK_IS_TOOLBAR(widget->parent);
}

static gboolean is_list(GtkWidget *widget)
{
    return widget &&
           (GTK_IS_CLIST(widget) ||
            GTK_IS_LIST(widget) ||
#if GTK_MAJOR_VERSION>1
            GTK_IS_TREE_VIEW(widget) ||
#ifdef GTK_ENABLE_BROKEN
            GTK_IS_TREE(widget) ||
#endif
#else
            GTK_IS_TREE(widget) ||
#endif
            GTK_IS_CTREE(widget));

}

static gboolean is_list_view_header(GtkWidget *widget)
{
    return widget && GTK_IS_BUTTON(widget) && widget->parent && is_list(widget->parent);
}

#if GTK_MAJOR_VERSION>1
/* CPD Another HACK!!! */
struct _GtkRangeLayout
{
    GdkRectangle stepper_a,
                 stepper_b,
                 stepper_c,
                 stepper_d;
};

static gboolean is_stepper_a(GtkWidget * widget, gint x, gint y)
{
    if (GTK_IS_RANGE(widget))
    {
        GtkRange * range = GTK_RANGE(widget);
        return range->has_stepper_a && (range->layout->stepper_a.x == (x - widget->allocation.x))
                                    && (range->layout->stepper_a.y == (y - widget->allocation.y));
    }

    return FALSE;
}

#if 0
static gboolean is_stepper_b(GtkWidget * widget, gint x, gint y)
{
    if (GTK_IS_RANGE(widget))
    {   
        GtkRange *range = GTK_RANGE(widget);
        return range->has_stepper_b && (range->layout->stepper_b.x == (x - widget->allocation.x))
                                    && (range->layout->stepper_b.y == (y - widget->allocation.y));
    }

    return FALSE;
}

static gboolean is_stepper_c(GtkWidget * widget, gint x, gint y)
{
    if (GTK_IS_RANGE(widget))
    {   
        GtkRange *range = GTK_RANGE(widget);
        return range->has_stepper_c && (range->layout->stepper_c.x == (x - widget->allocation.x))
                                    && (range->layout->stepper_c.y == (y - widget->allocation.y));
    }

    return FALSE;
}
#endif

static gboolean is_stepper_d(GtkWidget * widget, gint x, gint y)
{
    if (GTK_IS_RANGE(widget))
    {   
        GtkRange *range = GTK_RANGE(widget);
        return range->has_stepper_d && (range->layout->stepper_d.x == (x - widget->allocation.x))
                                    && (range->layout->stepper_d.y == (y - widget->allocation.y));
    }

    return FALSE;
}

#endif

static ERound get_round(const char *detail, GtkWidget *widget
#if GTK_MAJOR_VERSION>1
                        , int x, int y
#endif
                       )
{
    if(global.rounded)
        if(detail)
            if(0==strcmp(detail, "slider"))
                return ROUNDED_NONE;
            else if(0==strcmp(detail, "splitter") || 0==strcmp(detail, "optionmenu")  ||
                    0==strcmp(detail, "togglebutton") || 0==strcmp(detail, "hscale") || 0==strcmp(detail, "vscale")
                    /* || 0==strcmp(detail, "paned") || 0==strcmp(detail, QTC_PANED)*/ )
                return ROUNDED_ALL;
            else if(0==strcmp(detail, "spinbutton_up"))
                return ROUNDED_TOP;
            else if(0==strcmp(detail, "spinbutton_down"))
                return ROUNDED_BOTTOM;
            else if(0==strcmp(detail, "vscrollbar") || 0==strcmp(detail, "hscrollbar"))
            {
#if GTK_MAJOR_VERSION>1
                return is_stepper_a(widget, x, y)
                           ? 'h'==detail[0]
                               ? ROUNDED_LEFT
                               : ROUNDED_TOP
                           : is_stepper_d(widget, x, y)
                               ? 'v'==detail[0]
                                   ? ROUNDED_BOTTOM
                                   : ROUNDED_RIGHT
                               : ROUNDED_NONE;
#endif
                return ROUNDED_NONE;
            }
            else if(0==strcmp(detail, "button"))
            {
                if(is_list_view_header(widget))
                    return ROUNDED_NONE;
                else
                    return ROUNDED_ALL;
            }

    return ROUNDED_NONE;
}

static GdkGC * get_parent_bg_gc(GtkWidget *widget)
{
    if(GTK_IS_SCROLLBAR(widget))
        widget=widget->parent;

    return widget && widget->parent && widget->parent->style
               ? widget->parent->style->bg_gc[widget->parent->state]
               : NULL;
}

static void draw_bgnd(GdkWindow *window, GdkGC *gc, GtkWidget *widget, GdkRectangle *area, int x, int y, int width,
                     int height)
{
    GdkGC *parent_gc=get_parent_bg_gc(widget),
          *bgnd_gc=global.rounded && parent_gc
                       ? parent_gc
                       : gc;

    if(area)
        gdk_gc_set_clip_rectangle(bgnd_gc, area);
    gdk_draw_rectangle(window, bgnd_gc, TRUE, x, y, width, height);
    if(area)
        gdk_gc_set_clip_rectangle(bgnd_gc, NULL);
}

static void rgb_to_hls(gdouble *r, gdouble *g, gdouble *b)
{
    gdouble min,
            max,
            red=*r,
            green=*g,
            blue=*b,
            h,
            l, 
            s,
            delta;
  
    if(red > green)
    {
        max=MAX(red, blue);
        min=MIN(green, blue);
    }
    else
    {
        max=MAX(green, blue);
        min=MIN(red, blue);
    }
  
    l =(max + min) / 2;
    s = 0;
    h = 0;
  
    if(max != min)
    {
        if(l <= 0.5)
            s =(max - min) /(max + min);
        else
            s =(max - min) /(2 - max - min);
      
        delta = max -min;
        if(red == max)
            h =(green - blue) / delta;
        else if(green == max)
            h = 2 +(blue - red) / delta;
        else if(blue == max)
            h = 4 +(red - green) / delta;
      
        h *= 60;
        if(h < 0.0)
            h += 360;
    }
  
    *r = h;
    *g = l;
    *b = s;
}

static void hls_to_rgb(gdouble *h, gdouble *l, gdouble *s)
{
    gdouble hue,
            lightness=*l,
            saturation=*s,
            m1, 
            m2,
            r, 
            g, 
            b;

    if(lightness <= 0.5)
        m2 = lightness *(1 + saturation);
    else
        m2 = lightness + saturation - lightness * saturation;
    m1 = 2 * lightness - m2;
  
    if(saturation == 0)
    {
        *h = lightness;
        *l = lightness;
        *s = lightness;
    }
    else
    {
        hue = *h + 120;
        while(hue > 360)
            hue -= 360;
        while(hue < 0)
            hue += 360;
      
        if(hue < 60)
            r = m1 +(m2 - m1) * hue / 60;
        else if(hue < 180)
            r = m2;
        else if(hue < 240)
            r = m1 +(m2 - m1) *(240 - hue) / 60;
        else
            r = m1;
      
        hue = *h;
        while(hue > 360)
            hue -= 360;
        while(hue < 0)
            hue += 360;
      
        if(hue < 60)
            g = m1 +(m2 - m1) * hue / 60;
        else if(hue < 180)
            g = m2;
        else if(hue < 240)
            g = m1 +(m2 - m1) *(240 - hue) / 60;
        else
            g = m1;
      
        hue = *h - 120;
        while(hue > 360)
            hue -= 360;
        while(hue < 0)
            hue += 360;
      
        if(hue < 60)
            b = m1 +(m2 - m1) * hue / 60;
        else if(hue < 180)
            b = m2;
        else if(hue < 240)
            b = m1 +(m2 - m1) *(240 - hue) / 60;
        else
            b = m1;
      
        *h = r;
        *l = g;
        *s = b;
    }
}

static void shade(GdkColor * a, GdkColor * b, float k)
{
    gdouble red =(gdouble) a->red / 65535.0,
            green =(gdouble) a->green / 65535.0,
            blue =(gdouble) a->blue / 65535.0;
  
    rgb_to_hls(&red, &green, &blue);
  
    green *= k;
    if(green > 1.0)
        green = 1.0;
    else if(green < 0.0)
        green = 0.0;
  
    blue *= k;
    if(blue > 1.0)
        blue = 1.0;
    else if(blue < 0.0)
        blue = 0.0;
  
    hls_to_rgb(&red, &green, &blue);
  
    b->red = red * 65535.0;
    b->green = green * 65535.0;
    b->blue = blue * 65535.0;
}

static void draw_gradient(GdkWindow *window, GdkGC *gc, GdkColormap *colormap, GdkRectangle *area, GdkRegion *region,
                          int x, int y, int width, int height, GdkColor *begin_color, GdkColor *end_color,
                          gboolean horiz, gboolean increase)
{
    if(width>0 && height>0)
    {
        GdkColor     col=*begin_color;
        int          i,
                     last=horiz ? height : width,
                     dr=(end_color->red - begin_color->red) / last,
                     dg=(end_color->green - begin_color->green) / last,
                     db=(end_color->blue - begin_color->blue) / last;
        GdkRectangle clip;
        GdkGCValues  old_values;

        clip.x = x;
        clip.y = y;
        clip.width = width;
        clip.height = height;

        gdk_gc_get_values(gc, &old_values);

        if(area)
        {
            GdkRectangle clip2;

            if(gdk_rectangle_intersect(area, &clip, &clip2))
                gdk_gc_set_clip_rectangle(gc, gdk_rectangle_intersect(area, &clip, &clip2) ? &clip2 : &clip);
        }
        else if(region)
            gdk_gc_set_clip_region(gc, region);

        if(increase)
            for(i = 0; i < last; i++)
            {
                gdk_rgb_find_color(colormap, &col);
                gdk_gc_set_foreground(gc, &col);
                if(horiz)
                    gdk_draw_line(window, gc, x, y + i, x + width - 1, y + i);
                else
                    gdk_draw_line(window, gc, x + i, y,  x + i, y + height - 1);
                col.red += dr;
                col.green += dg;
                col.blue += db;
            }
        else
            for(i=last-1; i>=0; i--)
            {
                gdk_rgb_find_color(colormap, &col);
                gdk_gc_set_foreground(gc, &col);
                if(horiz)
                    gdk_draw_line(window, gc, x, y + i, x + width - 1, y + i);
                else
                    gdk_draw_line(window, gc, x + i, y,  x + i, y + height - 1);
                col.red += dr;
                col.green += dg;
                col.blue += db;
            }

        gdk_gc_set_foreground(gc, &old_values.foreground);
        if(area)
            gdk_gc_set_clip_rectangle(gc, NULL);
        else if(region)
            gdk_gc_set_clip_region(gc, NULL);
    }
}

static int equal(double d1, double d2)
{
    return (fabs(d1 - d2) < 0.0001);
}

static void draw_bevel_gradient(GdkWindow *window, GdkGC *gc, GdkColormap *colormap, GdkRectangle *area, 
                                GdkRegion *region, int x, int y, int width, int height, GdkColor *base,
                                double shade_top, double shade_bot, gboolean horiz, gboolean increase)
{
/*CPD TODO: Store last settings to make faster! */
    GdkColor top,
             bot,
             *t,
             *b;

    top.pixel=bot.pixel=0;

    if(equal(1.0, shade_top))
        t=base;
    else
    {
        shade(base, &top, shade_top);
        t=&top;
    }
    if(equal(1.0, shade_bot))
        b=base;
    else
    {
        shade(base, &bot, shade_bot);
        b=&bot;
    }

    draw_gradient(window, gc, colormap, area, region, x, y, width, height, t, b, horiz, increase);
}

static void draw_lines(GdkWindow *window, int rx, int ry, int rwidth, int rheight, gboolean horiz, int n_lines,
                       int offset, GdkGC *gcs[], GdkRectangle *area, gboolean etched)
{
    int   space =(n_lines*2)+(n_lines-1),
          x = horiz ? rx : rx+((rwidth-space)>>1),
          y = horiz ? ry+((rheight-space)>>1) : ry,
          x2 = rx + rwidth-1,
          y2 = ry + rheight-1,
          i,
          displacement = etched ? 1 : 0;
    GdkGC *gc1 = gcs[etched ? 4 : 0],
          *gc2 = gcs[etched ? 0 : 4];

    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, area);
        gdk_gc_set_clip_rectangle(gc2, area);
    }
    if(horiz)
    {
        for(i=0; i<space; i+=3)
            gdk_draw_line(window, gc1, x+offset, y+i, x2-(offset+displacement), y+i);
            
        for(i=1; i<space; i+=3)
             gdk_draw_line(window, gc2, x+offset+displacement, y+i, x2-offset, y+i);
    }   
    else
    {
        for(i=0; i<space; i+=3)
            gdk_draw_line(window, gc1, x+i, y+offset, x+i, y2-(offset+displacement));

        for(i=1; i<space; i+=3)
            gdk_draw_line(window, gc2, x+i, y+offset+displacement, x+i, y2-offset);
    }
    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, NULL);
        gdk_gc_set_clip_rectangle(gc2, NULL);
    }
}

static void draw_check(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                       GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                       gint height)
{
    QT_CURVE_STYLE_VAR
    gboolean mnu=DETAIL("check"),
             list=!mnu && is_list(widget),
             on=GTK_SHADOW_IN==shadow_type;

    x+=(width-QTC_CHECK_SIZE)>>1;
    y+=(height-QTC_CHECK_SIZE)>>1;

#if 0
printf("Draw check %d %s  ", state, detail ? detail : "NULL");
display_widget(widget, 3);
#endif

    if(list)
    {
        if(area)
            gdk_gc_set_clip_rectangle(style->text_gc[state], area);

        x++; y++;
        gdk_draw_rectangle(window, style->text_gc[state], FALSE, x, y, QTC_CHECK_SIZE-1, QTC_CHECK_SIZE-1);
        gdk_draw_rectangle(window, style->text_gc[state], FALSE, x+1, y+1, QTC_CHECK_SIZE-3, QTC_CHECK_SIZE-3);

        if(area)
            gdk_gc_set_clip_rectangle(style->text_gc[state], NULL);
    }
    else if(!mnu)
    {
        GdkGC    *new_gcs[7],
                 **btn_gcs=NULL;
        GdkColor new_colors[7],
                 *btn_colors;
        gboolean custom_c = FALSE;

        if(QT_CUSTOM_COLOR_BUTTON(style))
        {
            custom_c=TRUE;
            shade_colors(&(style->bg[state]), new_colors);
            GEN_GCS(style, new_colors, new_gcs);
            btn_gcs=new_gcs;
            btn_colors=new_colors;
        }
        else
        {
            btn_colors=qtcurve_style->button;
            btn_gcs=qtcurve_style->button_gc;
        }

        if(area)
        {
            gdk_gc_set_clip_rectangle(style->base_gc[state], area);
            gdk_gc_set_clip_rectangle(btn_gcs[5], area);
            gdk_gc_set_clip_rectangle(btn_gcs[4], area);
            gdk_gc_set_clip_rectangle(btn_gcs[1], area);
        }

        if(global.gradient_effect)
        {
            int bgnd=GTK_STATE_INSENSITIVE==state
                         ? 0
                         : GTK_STATE_PRELIGHT==state
                             ? on
                                 ? 2
                                 : 1
                             : on
                                 ? 3
                                 : 6;

            height=width=QTC_CHECK_SIZE;

            if(!global.border_button)
            {
                x--; y--; width+=2; height+=2;
            }

            draw_bevel_gradient(window, btn_gcs[bgnd], style->colormap, area, NULL, x+1, y+1, width-2, height-2,
                                &qtcurve_style->button[bgnd], 
                                on ? SHADE_BEVEL_GRAD_SEL_LIGHT : SHADE_BEVEL_GRAD_LIGHT,
                                on ? SHADE_BEVEL_GRAD_SEL_DARK : SHADE_BEVEL_GRAD_DARK, TRUE, !on);

            if(!on || !global.border_button)
            {
                int c1=on ? !global.border_button ? 5 : 4 : 0,
                    c2=on ? 0 : !global.border_button ? 5 : 4;

                gdk_draw_line(window, btn_gcs[c1], x+1, y+height-2, x+1, y+1);
                gdk_draw_line(window, btn_gcs[c1], x+1, y+1, x+width-2, y+1);
                gdk_draw_line(window, btn_gcs[c2], x+width-2, y+height-2, x+width-2, y+1);
                gdk_draw_line(window, btn_gcs[c2], x+1, y+height-2, x+width-2, y+height-2);
            }

            if(!global.border_button)
            {
                x++; y++; width-=2; height-=2;
            }
        }
        else
        {
            gdk_draw_rectangle(window, style->base_gc[state], TRUE, x+2, y+2, QTC_CHECK_SIZE-3, QTC_CHECK_SIZE-3);
            gdk_draw_line(window, btn_gcs[4], x+1, y+QTC_CHECK_SIZE-1, x+1, y+1);
            gdk_draw_line(window, btn_gcs[4], x+1, y+1, x+QTC_CHECK_SIZE-2, y+1);
        }

        if(global.rounded)
        {
            QTC_SET_MID_COLOR(btn_gcs[0], btn_colors[3], style->bg[state])

            gdk_draw_line(window, btn_gcs[5], x+1, y, x+width-2, y);
            gdk_draw_line(window, btn_gcs[5], x+1, y+height-1, x+width-2, y+height-1);
            gdk_draw_line(window, btn_gcs[5], x, y+1, x, y+height-2);
            gdk_draw_line(window, btn_gcs[5], x+width-1, y+1, x+width-1, y+height-2);

            gdk_draw_point(window, qtc_mid_color_gc, x, y);
            gdk_draw_point(window, qtc_mid_color_gc, x, y+width-1);
            gdk_draw_point(window, qtc_mid_color_gc, x+height-1, y);
            gdk_draw_point(window, qtc_mid_color_gc, x+height-1, y+width-1);
            QTC_UNSET_MID_COLOR
        }
        else if(!global.gradient_effect || global.border_button)
            gdk_draw_rectangle(window, btn_gcs[5], FALSE, x, y, QTC_CHECK_SIZE-1, QTC_CHECK_SIZE-1);

        if(custom_c)
            RELEASE_GCS(new_gcs);
    }

    if(on)
    {
        int      i;
        GdkPoint check[3]={ {x+3, y+4}, {x+5, y+6}, {x+9, y+2} }; /* , {x+9, y+5}, {x+5, y+9}, {x+3, y+7}}; */

        if(area)
            gdk_gc_set_clip_rectangle(style->text_gc[state], area);

        for(i=0; i<3; i++)
        {
            check[0].y++;
            check[1].y++;
            check[2].y++;
            gdk_draw_lines(window, style->text_gc[state], check, 3);
        }
        if(area)
            gdk_gc_set_clip_rectangle(style->text_gc[state], NULL);
    }

    if(!mnu && !list && area)
    {
        gdk_gc_set_clip_rectangle(style->base_gc[state], NULL);
        gdk_gc_set_clip_rectangle(qtcurve_style->button_gc[5], NULL);
        gdk_gc_set_clip_rectangle(qtcurve_style->button_gc[4], NULL);
        gdk_gc_set_clip_rectangle(qtcurve_style->button_gc[1], NULL);
    }
}

static void draw_option(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                        GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                        gint height)   
{
    QT_CURVE_STYLE_VAR
    gboolean mnu=DETAIL("option"),
             list=!mnu && is_list(widget);

    if(mnu)
        draw_check(style, window, state, shadow_type, area, widget, "check", x, y, width, height);
    else
    {
        gboolean set=GTK_SHADOW_IN==shadow_type;

        x+=(width-QTC_RADIO_SIZE)>>1;
        y+=(height-QTC_RADIO_SIZE)>>1;
            
        { /* So can modify X above, and then decalre outer... */
        
        GdkPoint outer[24]={ {x, y+8}, {x, y+4}, {x+1, y+3}, {x+1, y+2}, {x+2, y+1}, {x+3, y+1},
                        {x+4, y}, {x+8, y}, {x+9, y+1}, {x+10, y+1}, {x+11, y+2}, {x+11, y+3},
                        {x+12, y+4}, {x+12, y+8}, {x+11, y+9}, {x+11, y+10}, {x+10, y+11}, {x+9, y+11},
                        {x+8, y+12}, {x+4, y+12}, {x+3, y+11}, {x+2, y+11}, {x+1, y+10}, {x+1, y+9} };
                        
        if(list)
        {
            GdkPoint inner[20]={ {x+1,  y+8},  {x+1,  y+4},  {x+2,  y+3},  {x+2,  y+2},
                                 {x+3,  y+2},  {x+4,  y+1},  {x+8,  y+1},  {x+9,  y+2},
                                 {x+10, y+2},  {x+10, y+3},  {x+11, y+4},  {x+11, y+8},
                                 {x+10, y+9},  {x+10, y+10}, {x+9,  y+10}, {x+8,  y+11},
                                 {x+4,  y+11}, {x+3,  y+10}, {x+2,  y+10}, {x+2,  y+9} },
                     aa[16]=   { {x+2,  y+4}, {x+4, y+2},  {x+8, y+2},  {x+10, y+4},
                                 {x+10, y+8}, {x+8, y+10}, {x+4, y+10}, {x+2,  y+8},
                                 {x,    y+3}, {x+3, y},    {x+9, y},    {x+12, y+3},
                                 {x+12, y+9}, {x+9, y+12}, {x+3, y+12}, {x,    y+9} };

            QTC_SET_MID_COLOR_FACTOR(style->base_gc[state], style->base[state], style->text[state], 1.5)
            if(area)
                gdk_gc_set_clip_rectangle(style->text_gc[state], area);
            gdk_draw_points(window, style->text_gc[state], outer, 24);
            gdk_draw_points(window, style->text_gc[state], inner, 20);           
            gdk_draw_points(window, qtc_mid_color_gc, aa, 16);
            if(area)
                gdk_gc_set_clip_rectangle(style->text_gc[state], NULL);
            QTC_UNSET_MID_COLOR
        }
        else
        {
            GdkGC    *new_gcs[7],
                     **btn_gcs=NULL;
            GdkColor new_colors[7],
                     *btn_colors,
                     *left_shadow_color,
                     *right_shadow_color,
                     outer_left_color,
                     outer_right_color;
            gboolean custom_c = FALSE;
            int      bgnd=0;
            GdkPoint shadow[9]={ /*{x+4, y+11}, */{x+2, y+11}, {x+2, y+9}, {x+1, y+8}, {x+1, y+4},
                                  {x+2, y+3}, {x+2, y+2}, {x+3, y+2}, {x+4, y+1}, {x+8, y+1} },
                     outer_aa_left[8]= { { x,    y+3}, {x+1,  y+1},  {x+3,  y},     { x+9,  y},
                                        {x+11, y+1},  {x+12, y+3},  { x+1,  y+11}, {x,    y+9} },
                     outer_aa_right[4]={ { x+12, y+9}, {x+11, y+11}, {x+9,  y+12},  { x+3,  y+12} };
                     
            if(QT_CUSTOM_COLOR_BUTTON(style))
            {
                custom_c=TRUE;
                shade_colors(&(style->bg[state]), new_colors);
                GEN_GCS(style, new_colors, new_gcs);
                btn_gcs=new_gcs;
                btn_colors=new_colors;
            }
            else
            {
                btn_colors=qtcurve_style->button;
                btn_gcs=qtcurve_style->button_gc;
            }

            if(area)
            {
                gdk_gc_set_clip_rectangle(style->base_gc[state], area);
                gdk_gc_set_clip_rectangle(btn_gcs[5], area);
                gdk_gc_set_clip_rectangle(btn_gcs[4], area);
                gdk_gc_set_clip_rectangle(btn_gcs[0], area);
            }
    
            if(global.gradient_effect)
            {
                bgnd=GTK_STATE_INSENSITIVE==state
                        ? 0
                        : GTK_STATE_PRELIGHT==state
                            ? set
                                ? 2
                                : 1
                            : set
                                ? 3
                                : 6;
                GdkRegion *region=gdk_region_polygon(outer, 24, GDK_EVEN_ODD_RULE);
    
                draw_bevel_gradient(window, btn_gcs[bgnd], style->colormap, NULL, region, x+1, y+1, width-2, height-2,
                                    &qtcurve_style->button[bgnd], 
                                    set ? SHADE_BEVEL_GRAD_SEL_LIGHT : SHADE_BEVEL_GRAD_LIGHT,
                                    set ? SHADE_BEVEL_GRAD_SEL_DARK : SHADE_BEVEL_GRAD_DARK, TRUE, !set);
                gdk_region_destroy(region);
                if(!set || !global.border_button)
                {
                    if(!global.border_button)
                    {
                        GdkPoint shadow1b[14]= { {x+1, y+10}, {x+1, y+9},
                                                {x, y+8}, {x, y+4}, {x+1, y+3}, {x+1, y+2}, {x+2, y+1}, {x+3, y+1},
                                                {x+4, y}, {x+8, y}, {x+9, y+1}, {x+10, y+1}, {x+11, y+2}, {x+11, y+3} },
                                shadow2[10]= { {x+12, y+4}, {x+12, y+8}, {x+11, y+9}, {x+11, y+10}, {x+10, y+11},
                                                {x+9, y+11}, {x+8, y+12}, {x+4, y+12}, {x+3, y+11}, {x+2, y+11} } ;
                        gdk_draw_lines(window, btn_gcs[set ? 5 : 0], shadow1b, 14);
                        gdk_draw_lines(window, btn_gcs[set ? 0 : 5], shadow2, 10);
                        left_shadow_color=&(btn_colors[5]);
                    }
                    else
                    {
                        GdkPoint shadow2[9]={ {x+10, y+2},  {x+10, y+3}, {x+11, y+4}, {x+11, y+8}, {x+10, y+9}, 
                                            {x+10, y+10}, {x+9, y+10}, {x+8, y+11}, {x+4, y+11} };
    
                        gdk_draw_lines(window, btn_gcs[set ? 4 : 0], shadow, 9);
                        gdk_draw_lines(window, btn_gcs[set ? 0 : 4], shadow2, 9);
                        left_shadow_color=&(btn_colors[4]);
                    }
                    right_shadow_color=&(btn_colors[set ? 0 : !global.border_button ? 5 : 4]);
                }
                else
                    left_shadow_color=right_shadow_color=&(btn_colors[bgnd]);
            }
            else
            {
                gdk_draw_rectangle(window, style->base_gc[state], TRUE, x+2, y+2, QTC_RADIO_SIZE-4, QTC_RADIO_SIZE-3);
                gdk_draw_line(window, style->base_gc[state], x+QTC_RADIO_SIZE-2, y+4, x+QTC_RADIO_SIZE-2, y+8);
                gdk_draw_lines(window, btn_gcs[4], shadow, 9);
                right_shadow_color=&(style->base[state]);
                left_shadow_color=&(btn_colors[4]);
            }
    
            if(!global.gradient_effect || global.border_button)
            {
                gdk_draw_lines(window, btn_gcs[5], outer, 24);
                shade(&btn_colors[5], &outer_right_color, 1.1);
            }
            else
            {
                shade(left_shadow_color, &outer_left_color, 1.1);
                shade(right_shadow_color, &outer_left_color, 1.1);
            }
            
            QTC_SET_MID_COLOR(btn_gcs[3], outer_right_color, style->bg[state])
            gdk_draw_points(window, qtc_mid_color_gc, outer_aa_right, 8);
            if(global.gradient_effect && !global.border_button)
            {
                QTC_UNSET_MID_COLOR
                QTC_SET_MID_COLOR(btn_gcs[3], outer_left_color, style->bg[state])
            }
            gdk_draw_points(window, qtc_mid_color_gc, outer_aa_left, 8);
            QTC_UNSET_MID_COLOR
            QTC_SET_MID_COLOR(btn_gcs[3], btn_colors[bgnd], btn_colors[5])
    
            if(global.gradient_effect)
            {
                if(!set)
                    if(global.border_button)
                    {
                        GdkPoint inner_aa[3]={ {x+1, y+4},  {x+2, y+2},  {x+4, y+1} };
                        gdk_draw_points(window, qtc_mid_color_gc, inner_aa, 3);
                        QTC_UNSET_MID_COLOR
                        QTC_SET_MID_COLOR(btn_gcs[3], outer_right_color, style->bg[state])
                        gdk_draw_point(window, qtc_mid_color_gc, x+2, y+10);
                    }
                    else
                    {
                        GdkPoint inner_aa[4]={ {x+4,  y+11},  {x+8,  y+11},  {x+10, y+10}, {x+11, y+8} };
                        gdk_draw_points(window, qtc_mid_color_gc, inner_aa, 4);
                    }
            }
            else
            {
                GdkPoint inner_aa[6]={ {x+4,  y+11},  {x+8,  y+11},  {x+10, y+10},
                                    {x+11, y+8},   {x+11, y+4},   {x+10, y+2} };
                gdk_draw_points(window, qtc_mid_color_gc, inner_aa, 6);
            }
            QTC_UNSET_MID_COLOR

            if(area)
            {
                gdk_gc_set_clip_rectangle(style->base_gc[state], NULL);
                gdk_gc_set_clip_rectangle(btn_gcs[5], NULL);
                gdk_gc_set_clip_rectangle(btn_gcs[4], NULL);
                gdk_gc_set_clip_rectangle(btn_gcs[0], NULL);
            }
            if(custom_c)
                RELEASE_GCS(new_gcs);
        }
        
        if(set)
        {
            QTC_SET_MID_COLOR(style->base_gc[state], style->base[state], style->text[state])

            if(area)
                gdk_gc_set_clip_rectangle(style->text_gc[state], area);

            gdk_draw_line(window, qtc_mid_color_gc, x+5, y+4, x+7, y+4);
            gdk_draw_line(window, qtc_mid_color_gc, x+5, y+8, x+7, y+8);
            gdk_draw_line(window, qtc_mid_color_gc, x+4, y+5, x+4, y+7);
            gdk_draw_line(window, qtc_mid_color_gc, x+8, y+5, x+8, y+7);
            gdk_draw_rectangle(window, style->text_gc[state], TRUE, x+5, y+5, 3, 3);

            if(area)
                gdk_gc_set_clip_rectangle(style->text_gc[state], NULL);

            QTC_UNSET_MID_COLOR
        }
        }   /* So can modify X above, and then decalre outer... */
    }
}

static void draw_box(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                     GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                     gint height);

static void draw_shadow(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                        GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                        gint height);
#if GTK_MAJOR_VERSION==1
static void real_draw_slider(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                             GdkRectangle *area, GtkWidget *widget, gchar *detail, gint x, gint y, gint width,
                             gint height, GtkOrientation orientation);
#endif

#define IS_MENU_ITEM(WIDGET)(WIDGET &&(GTK_IS_MENU_ITEM(WIDGET) ||(WIDGET->parent && GTK_IS_MENU_ITEM(WIDGET->parent))))

static gboolean is_on_a_button(GtkWidget *w, int level, gboolean *def)
{
    if(w)
    {
        if(GTK_IS_BUTTON(w) &&(!(GTK_IS_RADIO_BUTTON(w) || GTK_IS_CHECK_BUTTON(w) || GTK_IS_OPTION_MENU(w))))
        {
            if(def)
                *def=GTK_WIDGET_HAS_DEFAULT(w);
            return TRUE;
        }
        else if(level<3)
            return is_on_a_button(w->parent, level++, def);
    }

    return FALSE;
}   

#if GTK_MAJOR_VERSION>1
static void draw_layout(GtkStyle *style, GdkWindow *window, GtkStateType state, gboolean use_text,
                        GdkRectangle *area, GtkWidget *widget, const gchar *detail, gint x, gint y,
                        PangoLayout *layout)
{
    gboolean selectedText=IS_MENU_ITEM(widget) && GTK_STATE_PRELIGHT==state,
             def_but=FALSE;

#if 0
printf("Draw layout %s %d %d %s  ", pango_layout_get_text(layout), state, use_text, detail ? detail : "NULL");
display_widget(widget, 3);
#endif

    /* The following fixes the text in list views... if not used, when an item is selected it gets the selected
       text colour - but when the window changes focus it gets the normal text colour! */
    if(DETAIL("cellrenderertext") && GTK_STATE_ACTIVE==state)
        state=GTK_STATE_SELECTED;

    if(global.bold_def_text)  /* Only need to check if this style option is set */
        is_on_a_button(widget, 0, &def_but);

    parent_class->draw_layout(style, window, selectedText ? GTK_STATE_SELECTED : state, use_text || selectedText,
                              area, widget, detail, x, y, layout);
    if(def_but)
        parent_class->draw_layout(style, window, selectedText ? GTK_STATE_SELECTED : state, use_text || selectedText,
                                  area, widget, detail, x+1, y, layout);
}
#else

static void draw_string(GtkStyle *style, GdkWindow *window, GtkStateType state, GdkRectangle *area,
                        GtkWidget *widget, PCONST gchar *detail, gint x, gint y, const gchar * string)
{
    gboolean def_but;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

#if 0
printf("Draw string %d %s \"%s\" %d :%d: ", state, detail ? detail : "NULL", string ? string : "NULL",
widget && widget->parent ? GTK_WIDGET_STATE(widget->parent) : -1,
is_on_a_button(widget, 0, &def_but));
display_widget(widget, 3);
#endif

    if(is_on_a_button(widget, 0, &def_but))
    {
        GdkGC *gc=GTK_STATE_PRELIGHT==state || GTK_STATE_ACTIVE==state
                      ? style->text_gc[GTK_STATE_NORMAL]
                      : style->text_gc[state];
        int   x_offset=GTK_STATE_ACTIVE==state ? 1 : 0,
              y_offset=x_offset; 

        if(area)
        {
            gdk_gc_set_clip_rectangle(style->white_gc, area);
            gdk_gc_set_clip_rectangle(gc, area);
        }
        if(GTK_STATE_INSENSITIVE==state)
        {
            gdk_draw_string(window, style->font, style->white_gc, x + 1 + x_offset, y + 1 + y_offset, string);
            if(global.bold_def_text && def_but)
                gdk_draw_string(window, style->font, style->white_gc, x + 1 + x_offset +1, y + 1 + y_offset, string);
        }
        gdk_draw_string(window, style->font, gc, x + x_offset, y + y_offset, string);
        if(global.bold_def_text && def_but)
            gdk_draw_string(window, style->font, gc, x + x_offset +1, y + y_offset, string);

        if(area)
        {
            gdk_gc_set_clip_rectangle(style->white_gc, NULL);
            gdk_gc_set_clip_rectangle(gc, NULL);
        }
    }
    else
        parent_class->draw_string(style, window, IS_MENU_ITEM(widget) && GTK_STATE_PRELIGHT==state ? GTK_STATE_SELECTED :
state,
                                  area, widget, detail, x, y, string);
}
#endif

static gboolean sanitize_size(GdkWindow *window, gint *width, gint *height)
{
    gboolean set_bg = FALSE;

    if((-1==*width) && (-1==*height))
    {
#if GTK_MAJOR_VERSION>1
        set_bg = GDK_IS_WINDOW(window);
#endif
        gdk_window_get_size(window, width, height);
    }
    else if(-1==*width)
        gdk_window_get_size(window, width, NULL);
    else if(-1==*height)
        gdk_window_get_size(window, NULL, height);

  return set_bg;
}

static void option_menu_get_props(GtkWidget *widget, GtkRequisition *indicator_size, GtkBorder *indicator_spacing)
{
    GtkRequisition *tmp_size = NULL;
    GtkBorder      *tmp_spacing = NULL;

#if GTK_MAJOR_VERSION>1
    if(widget)
        gtk_widget_style_get(widget, "indicator_size", &tmp_size, "indicator_spacing", &tmp_spacing, NULL);
#endif
    if(tmp_size)
    {
        *indicator_size = *tmp_size;
        g_free(tmp_size);
    }
    else
        *indicator_size = default_option_indicator_size;

    if(tmp_spacing)
    {
        *indicator_spacing = *tmp_spacing;
        g_free(tmp_spacing);
    }
    else
        *indicator_spacing = default_option_indicator_spacing;
}

static void arrow_draw_hline(GdkWindow *window, GdkGC *gc, int x1, int x2, int y, gboolean last)
{
    if(x2 - x1 < 7 && !last) /* 7 to get garretts pixels, otherwise 6 */
        gdk_draw_line(window, gc, x1, y, x2, y);
    else if(last)
    {
        if(x2 - x1 <= 7)
        {
            gdk_draw_line(window, gc, x1+1, y, x1+1, y);
            gdk_draw_line(window, gc, x2-1, y, x2-1, y); 
        }
        else
        {
            gdk_draw_line(window, gc, x1+2, y, x1+2, y);
            gdk_draw_line(window, gc, x2-2, y, x2-2, y);
        }
    }
    else
    {
        gdk_draw_line(window, gc, x1, y, x1+2, y);
        gdk_draw_line(window, gc, x2-2, y, x2, y); 
    }
}

static void arrow_draw_vline(GdkWindow *window, GdkGC *gc, int y1, int y2, int x, gboolean last)
{
    if(y2 - y1 < 7 && !last) /* 7 to get garretts pixels */
        gdk_draw_line(window, gc, x, y1, x, y2);
    else if(last)
    {
        gdk_draw_line(window, gc, x, y1+2, x, y1+2);
        gdk_draw_line(window, gc, x, y2-2, x, y2-2); 
    }
    else
    {
        gdk_draw_line(window, gc, x, y1, x, y1+2);
        gdk_draw_line(window, gc, x, y2-2, x, y2); 
    }
}

static void draw_arrow(GdkWindow *window, GdkGC *gc, GdkRectangle *area, GtkArrowType arrow_type, gint x, gint y,
                       gint width, gint height)
{
    gint i,
         j;

    if(area)
        gdk_gc_set_clip_rectangle(gc, area);

    if(global.v_arrow)
        switch(arrow_type)
        {
            case GTK_ARROW_DOWN:
                for(i = 0, j = -1; i < height; i++, j++)
                    arrow_draw_hline(window, gc, x + j, x + width - j - 1, y + i, i == 0);
                break;
            case GTK_ARROW_UP:
                for(i = height - 1, j = -1; i >= 0; i--, j++)
                    arrow_draw_hline(window, gc, x + j, x + width - j - 1, y + i, i == height - 1);
                break;
            case GTK_ARROW_LEFT:
                for(i = width - 1, j = -1; i >= 0; i--, j++)
                    arrow_draw_vline(window, gc, y + j, y + height - j - 1, x + i, i == width - 1);
                break;
            case GTK_ARROW_RIGHT:
                for(i = 0, j = -1; i < width; i++, j++)
                    arrow_draw_vline(window, gc, y + j, y + height - j - 1,  x + i, i == 0);
        }
    else
        switch(arrow_type)
        {
            case GTK_ARROW_DOWN:
                for(i = 0, j = 0; i < height; i++, j++)
                    gdk_draw_line(window, gc, x + j, y + i, x + width - j - 1, y + i);
                break;
            case GTK_ARROW_UP:
                for(i = height - 1, j = 0; i >= 0; i--, j++)
                    gdk_draw_line(window, gc, x + j, y + i, x + width - j - 1, y + i);
                break;
            case GTK_ARROW_LEFT:
                for(i = width - 1, j = 0; i >= 0; i--, j++)
                    gdk_draw_line(window, gc, x + i, y + j, x + i, y + height - j - 1);
                break;
            case GTK_ARROW_RIGHT:
                for(i = 0, j = 0; i < width; i++, j++)
                    gdk_draw_line(window, gc, x + i, y + j, x + i, y + height - j - 1);
        }

    if(area)
        gdk_gc_set_clip_rectangle(gc, NULL);
}

static void qtcurve_draw_arrow(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow,
                               GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, GtkArrowType arrow_type,
                               gboolean fill, gint x, gint y, gint width, gint height)
{
    QT_CURVE_STYLE_VAR

    if(DETAIL("arrow")) 
    {
        if(GTK_ARROW_DOWN==arrow_type || GTK_ARROW_UP==arrow_type)
            draw_arrow(window, style->text_gc[QTC_ARROW_STATE(state)], area,  arrow_type, x+((width - LARGE_ARR_WIDTH)>>1),
                       y+((height-LARGE_ARR_HEIGHT)>>1), LARGE_ARR_WIDTH, LARGE_ARR_HEIGHT);
        else
            draw_arrow(window, style->text_gc[QTC_ARROW_STATE(state)], area,  arrow_type,
                       x+((width - LARGE_ARR_HEIGHT)>>1),
                       y+((height-LARGE_ARR_WIDTH)>>1), LARGE_ARR_HEIGHT, LARGE_ARR_WIDTH); 
    }      
    else
    {
        int is_spinbutton = DETAIL("spinbutton"),
            a_width=LARGE_ARR_WIDTH,
            a_height=LARGE_ARR_HEIGHT;

        sanitize_size(window, &width, &height);

#if GTK_MAJOR_VERSION==1 
        if(DETAIL("hscrollbar") || DETAIL("vscrollbar") || is_spinbutton)
        {
            const char *box_detail = detail;
            int        box_x = x,
                       box_y = y,
                       box_width = width,
                       box_height = height;
      
            if(is_spinbutton)
            {
                box_detail = (GTK_ARROW_DOWN==arrow_type) ? "spinbutton_down" : "spinbutton_up";
                box_x -= 2;
                box_width += 4;
                if(GTK_ARROW_UP==arrow_type)
                {
                    box_y -= 2;
                    box_height += 2;
                }
                else
                {
                    box_y -= 0;
                    box_height += 1; /* 2; CPD: This worked before as 2 - maybe because new Gtk ver, which one!!!*/
                }
            }

            if(is_spinbutton && GTK_SHADOW_ETCHED_IN==shadow)
            {
                shadow = GTK_SHADOW_OUT;
                state = GTK_STATE_INSENSITIVE;
            }

            draw_box(style, window, state, shadow, area, widget, box_detail, box_x, box_y, box_width, box_height);

            if(is_spinbutton)
            {
                x += 3;
                y += 2;
                width -= 6;
                height -= 4;
            }
            else /* scrollbar */
            {
                x += 4;
                y += 4;
                width -= 8;
                height -= 8;
            }
        }
#else
        if(is_spinbutton)
            if(GTK_ARROW_UP==arrow_type)
                y-=1;
            else if(!global.v_arrow && GTK_ARROW_DOWN==arrow_type)
                y+=1;
#endif
        if(is_spinbutton)
        {
            a_height = SMALL_ARR_HEIGHT;
            a_width = SMALL_ARR_WIDTH;
        }
        else if(GTK_ARROW_LEFT==arrow_type || GTK_ARROW_RIGHT==arrow_type || DETAIL("menuitem"))
        {
            a_width = LARGE_ARR_HEIGHT;
            a_height = LARGE_ARR_WIDTH;
        }

        x+=(width-a_width)>>1;
        y+=(height-a_height)>>1;

        if(GTK_ARROW_RIGHT==arrow_type && (width-a_width)%2)
            x++;

        if(GTK_ARROW_DOWN==arrow_type && (height-a_height)%2)
            y++;

        draw_arrow(window, style->text_gc[IS_MENU_ITEM(widget) && GTK_STATE_PRELIGHT==state
                           ? GTK_STATE_SELECTED : QTC_ARROW_STATE(state)], 
                   area, arrow_type, x, y, a_width, a_height);
    }
}

static void draw_tab(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                     GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                     gint height)
{
    draw_arrow(window, style->text_gc[QTC_ARROW_STATE(state)], area, GTK_ARROW_DOWN, x+((width-LARGE_ARR_WIDTH)>>1),
               y+((height-LARGE_ARR_HEIGHT)>>1), LARGE_ARR_WIDTH, LARGE_ARR_HEIGHT);
}

static void draw_shadow(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                        GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                        gint height)
{
    QT_CURVE_STYLE_VAR
    gboolean use_bc = use_button_color(detail),
             custom_c = FALSE,
             frame=!detail || 0==strcmp(detail, "frame") || 0==strcmp(detail, "entry"),
             menu=DETAIL("menu");
    int      dark=(global.border_frame && menu) || (use_bc && global.border_button) ? 4 : 5;
    GdkGC    *outer_gc = NULL,
             *gc1 = NULL,
             *gc2 = NULL;
    ERound   round=get_round(detail, widget
#if GTK_MAJOR_VERSION>1
                             , x, y
#endif
                            );
    GdkGC    *new_gcs[7],
             **btn_gcs=NULL;
    GdkColor new_colors[7],
             *btn_colors;

#if 0
printf("Draw shadow %d %d %d %d %d %d %d %s  ", round, state, shadow_type, x, y, width, height, detail ? detail : "NULL");
display_widget(widget, 3);
#endif

#if GTK_MAJOR_VERSION==1
    if(detail && (0==strcmp(detail, "entry") || 0==strcmp(detail, "text")) && GTK_WIDGET_HAS_FOCUS(widget))
    {
        x--;
        y--;
        width+=2;
        height+=2;
    }
#endif

    if(use_bc && QT_CUSTOM_COLOR_BUTTON(style))
    {
        custom_c=TRUE;
        shade_colors(&(style->bg[state]), new_colors);
        GEN_GCS(style, new_colors, new_gcs);
        btn_gcs=new_gcs;
        btn_colors=new_colors;
    }
    else
    {
        btn_colors=qtcurve_style->button;
        btn_gcs=qtcurve_style->button_gc;
    }

    sanitize_size(window, &width, &height);
 
/* CPD - THIS WAS IN GTK2 VER, and NOT GTK1 ??? */
#if GTK_MAJOR_VERSION>1 && GTK_MINOR_VERSION<2
    if(DETAIL("slider") && widget && GTK_IS_RANGE(widget))
    {
        GtkAdjustment *adj = GTK_RANGE(widget)->adjustment;

        if(adj->value <= adj->lower && (GTK_RANGE(widget)->has_stepper_a || GTK_RANGE(widget)->has_stepper_b))
        {
            if(GTK_IS_VSCROLLBAR(widget))
            {
                height += 1;
                y -= 1;
            }
            else if(GTK_IS_HSCROLLBAR(widget))
            {
                width += 1;
                x -= 1;
            }
        }
        if(adj->value >= adj->upper - adj->page_size && (GTK_RANGE(widget)->has_stepper_c || 
           GTK_RANGE(widget)->has_stepper_d))
            if(GTK_IS_VSCROLLBAR(widget))
                height += 1;
            else if(GTK_IS_HSCROLLBAR(widget))
                width += 1;
    }
#endif
    if(DETAIL("menuitem"))
        outer_gc = qtcurve_style->gradient_gc[GRADIENT_TOP];
    else
        outer_gc = use_bc ? btn_gcs[5] : qtcurve_style->gray_gc[5];

    switch(shadow_type)
    {
        case GTK_SHADOW_NONE:
            return;
        case GTK_SHADOW_IN:
        case GTK_SHADOW_ETCHED_IN:
            if(frame)
            {
                gc1 = qtcurve_style->gray_gc[0];
                gc2 = qtcurve_style->gray_gc[dark];
            }
            else
            {
                gc1 = use_bc ? btn_gcs[0] : qtcurve_style->gray_gc[0];
                gc2 = use_bc ? btn_gcs[dark] : qtcurve_style->gray_gc[dark];
            }
            break;
        case GTK_SHADOW_OUT:
        case GTK_SHADOW_ETCHED_OUT:
            if(DETAIL("menuitem"))
            {
                gc1 = qtcurve_style->gradient_gc[GRADIENT_DARK];
                gc2 = qtcurve_style->gradient_gc[GRADIENT_LIGHT];
            }
            else if(!global.border_frame && (frame || menu))
            {
                gc1 = qtcurve_style->gray_gc[5];
                gc2 = qtcurve_style->gray_gc[0];
            }
            else
            {
                gc1 = use_bc ? btn_gcs[dark] : qtcurve_style->gray_gc[dark];
                gc2 = use_bc ? btn_gcs[0] : qtcurve_style->gray_gc[0];
            }
            break;
    }
  
    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, area);
        gdk_gc_set_clip_rectangle(gc2, area);
        if((GTK_SHADOW_IN==shadow_type ||  GTK_SHADOW_OUT==shadow_type) && outer_gc!=gc1 && outer_gc!=gc2)
            gdk_gc_set_clip_rectangle(outer_gc, area);
    }
  
    switch(shadow_type)
    {
        case GTK_SHADOW_NONE:
            break;
        case GTK_SHADOW_IN:
        case GTK_SHADOW_OUT:
            if(frame || menu)
            {
                if(menu && global.border_frame)
                {
                    gdk_draw_rectangle(window, outer_gc, FALSE, x, y, width - 1, height - 1);
                    x++; y++; width-=2; height-=2;
                }
        
                gdk_draw_line(window, gc2, x,         y,          x+width-2, y);
                gdk_draw_line(window, gc2, x,         y,          x,         y+height-2);
                gdk_draw_line(window, gc1, x,         y+height-1, x+width-1, y+height-1);
                gdk_draw_line(window, gc1, x+width-1, y+height-1, x+width-1, y);
                break;
            }

            if(!(GTK_SHADOW_IN==shadow_type && global.border_button) && detail &&
                    ((0==strcmp(detail, "button")) || 
                     (0==strcmp(detail, "togglebutton")) || 
                     (0==strcmp(detail, "handlebox")) ||
                     (0==strcmp(detail, "optionmenu")) ||
                     (0==strcmp(detail, "notebook")) ||
                     (0==strcmp(detail, "slider")) || 
                     (0==strcmp(detail, "spinbutton_up")) ||
                     (0==strcmp(detail, "spinbutton_down")) ||
                     (0==strcmp(detail, "menuitem")) ||
                     (0==strcmp(detail, "vscale")) ||
                     (0==strcmp(detail, "hscale")) ||
                     (0==strcmp(detail, "vscrollbar")) ||
                     (0==strcmp(detail, "hscrollbar")) ||
                     (0==strcmp(detail, "stepper")) ||
                     (0==strcmp(detail, QTC_PANED)) ))
            {
                if(!global.border_button) 
                {
                    x--; y--; width+=2; height+=2;
                }

                /* Top and left */
                if(QT_STYLE->ythickness > 0)
                    gdk_draw_line(window, gc2, x + 1, y + 1, x + width - 2, y + 1);
                if(QT_STYLE->xthickness > 0)
                    gdk_draw_line(window, gc2, x + 1, y + 1, x + 1, y + height - 2);

                /* Right and bottom edge */
                if(QT_STYLE->ythickness > 0)
                    gdk_draw_line(window, gc1, x + 1, y + height - 2, x + width - 2, y + height - 2);
                if(QT_STYLE->xthickness > 0)
                    gdk_draw_line(window, gc1, x + width - 2, y + 1, x + width - 2, y + height - 2);

                if(!global.border_button) 
                {
                    x++; y++; width-=2; height-=2;
                }
            }

            if(global.border_button)
                if(ROUNDED_NONE==round)
                    gdk_draw_rectangle(window, outer_gc, FALSE, x, y, width - 1, height - 1);
                else
                {
                    gboolean wide=width>=QTC_MIN_BTN_SIZE,
                             tall=height>=QTC_MIN_BTN_SIZE;
                    GdkGC    *gc=outer_gc;

                    QTC_SET_MID_COLOR(btn_gcs[4], btn_colors[5], style->bg[state])

                    switch(round)
                    {
                        case ROUNDED_ALL:
                            gdk_draw_line(window, gc, x+2, y, x+width-3, y);
                            gdk_draw_line(window, gc, x+2, y+height-1, x+width-3, y+height-1);
                            gdk_draw_line(window, gc, x, y+2, x, y+height-3);
                            gdk_draw_line(window, gc, x+width-1, y+2, x+width-1, y+height-3);
                            if(tall && wide)
                            {
                                gdk_draw_point(window, gc, x+1, y+1);
                                gdk_draw_point(window, gc, x+width-2, y+1);
                                gdk_draw_point(window, gc, x+1, y+height-2);
                                gdk_draw_point(window, gc, x+width-2, y+height-2);
                                gc=qtc_mid_color_gc;
                            }
                            gdk_draw_line(window, gc, x, y+1, x+1, y);
                            gdk_draw_line(window, gc, x+width-2, y, x+width-1, y+1);
                            gdk_draw_line(window, gc, x, y+height-2, x+1, y+height-1);
                            gdk_draw_line(window, gc, x+width-2, y+height-1, x+width-1, y+height-2);
                            if(!(tall && wide))
                            {
                                QTC_UNSET_MID_COLOR
                                QTC_SET_MID_COLOR(btn_gcs[4], btn_colors[3], style->bg[state])
                                gdk_draw_point(window, qtc_mid_color_gc, x, y);
                                gdk_draw_point(window, qtc_mid_color_gc, x+width-1, y);
                                gdk_draw_point(window, qtc_mid_color_gc, x, y+height-1);
                                gdk_draw_point(window, qtc_mid_color_gc, x+width-1, y+height-1);
                            }
                            break;
                        case ROUNDED_TOP:
                            gdk_draw_line(window, gc, x+2, y, x+width-3, y);
                            gdk_draw_line(window, gc, x+1, y+height-1, x+width-2, y+height-1);
                            gdk_draw_line(window, gc, x, y+2, x, y+height-1);
                            gdk_draw_line(window, gc, x+width-1, y+2, x+width-1, y+height-1);
                            if(wide)
                            {
                                gdk_draw_point(window, gc, x+1, y+1);
                                gdk_draw_point(window, gc, x+width-2, y+1);
                                gc=qtc_mid_color_gc;
                            }
                            gdk_draw_line(window, gc, x, y+1, x+1, y);
                            gdk_draw_line(window, gc, x+width-2, y, x+width-1, y+1);
                            if(!wide)
                            {
                                QTC_UNSET_MID_COLOR
                                QTC_SET_MID_COLOR(btn_gcs[4], btn_colors[3], style->bg[state])
                                gdk_draw_point(window, qtc_mid_color_gc, x, y);
                                gdk_draw_point(window, qtc_mid_color_gc, x+width-1, y);
                            }
                            break;
                        case ROUNDED_BOTTOM:
                            gdk_draw_line(window, gc, x+1, y, x+width-2, y);
                            gdk_draw_line(window, gc, x+2, y+height-1, x+width-3, y+height-1);
                            gdk_draw_line(window, gc, x, y, x, y+height-3);
                            gdk_draw_line(window, gc, x+width-1, y, x+width-1, y+height-3);
                            if(wide)
                            {
                                gdk_draw_point(window, gc, x+1, y+height-2);
                                gdk_draw_point(window, gc, x+width-2, y+height-2);
                                gc=qtc_mid_color_gc;
                            }
                            gdk_draw_line(window, gc, x, y+height-2, x+1, y+height-1);
                            gdk_draw_line(window, gc, x+width-2, y+height-1, x+width-1, y+height-2);
                            if(!wide)
                            {
                                QTC_UNSET_MID_COLOR
                                QTC_SET_MID_COLOR(btn_gcs[4], btn_colors[3], style->bg[state])
                                gdk_draw_point(window, qtc_mid_color_gc, x, y+height-1);
                                gdk_draw_point(window, qtc_mid_color_gc, x+width-1, y+height-1);
                            }
                            break;
                        case ROUNDED_LEFT:
                            gdk_draw_line(window, gc, x+2, y, x+width-2, y);
                            gdk_draw_line(window, gc, x+2, y+height-1, x+width-2, y+height-1);
                            gdk_draw_line(window, gc, x, y+2, x, y+height-3);
                            gdk_draw_line(window, gc, x+width-1, y, x+width-1, y+height-1);
                            if(tall)
                            {
                                gdk_draw_point(window, gc, x+1, y+1);
                                gdk_draw_point(window, gc, x+1, y+height-2);
                                gc=qtc_mid_color_gc;
                            }
                            gdk_draw_line(window, gc, x, y+1, x+1, y);
                            gdk_draw_line(window, gc, x, y+height-2, x+1, y+height-1);
                            if(!tall)
                            {
                                QTC_UNSET_MID_COLOR
                                QTC_SET_MID_COLOR(btn_gcs[4], btn_colors[3], style->bg[state])
                                gdk_draw_point(window, qtc_mid_color_gc, x, y);
                                gdk_draw_point(window, qtc_mid_color_gc, x, y+height-1);
                            }
                            break;
                        case ROUNDED_RIGHT:
                            gdk_draw_line(window, gc, x+1, y, x+width-3, y);
                            gdk_draw_line(window, gc, x+1, y+height-1, x+width-3, y+height-1);
                            gdk_draw_line(window, gc, x, y, x, y+height-1);
                            gdk_draw_line(window, gc, x+width-1, y+2, x+width-1, y+height-3);
                            if(tall)
                            {
                                gdk_draw_point(window, gc, x+width-2, y+1);
                                gdk_draw_point(window, gc, x+width-2, y+height-2);
                                gc=qtc_mid_color_gc;
                            }
                            gdk_draw_line(window, gc, x+width-2, y, x+width-1, y+1);
                            gdk_draw_line(window, gc, x+width-2, y+height-1, x+width-1, y+height-2);
                            if(!tall)
                            {
                                QTC_UNSET_MID_COLOR
                                QTC_SET_MID_COLOR(btn_gcs[4], btn_colors[3], style->bg[state])
                                gdk_draw_point(window, qtc_mid_color_gc, x+width-1, y);
                                gdk_draw_point(window, qtc_mid_color_gc, x+width-1, y+height-1);
                            }
                            break;
                        default:
                            break;
                    }
                    QTC_UNSET_MID_COLOR
                }
            break;
        case GTK_SHADOW_ETCHED_IN:
            gdk_draw_rectangle(window, gc1, FALSE, x+1, y+1, width-2, height-2);
            gdk_draw_rectangle(window, gc2, FALSE, x, y, width-2, height-2);
            break;
        case GTK_SHADOW_ETCHED_OUT:
            gdk_draw_rectangle(window, gc2, FALSE, x+1, y+1, width-2, height-2);
            gdk_draw_rectangle(window, gc1, FALSE, x, y, width-2, height-2);
            break;
    }

    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, NULL);
        gdk_gc_set_clip_rectangle(gc2, NULL);
        if((GTK_SHADOW_IN==shadow_type || GTK_SHADOW_OUT==shadow_type) && outer_gc!=gc1 && outer_gc!=gc2)
            gdk_gc_set_clip_rectangle(outer_gc, NULL);
    }

    if(custom_c)
        RELEASE_GCS(new_gcs);
}

static void draw_box_gap(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                         GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                         gint height, GtkPositionType gap_side, gint gap_x, gint gap_width)
{
    QT_CURVE_STYLE_VAR
    GdkGC *gc1 = NULL,
          *gc2 = NULL,
          *outer = qtcurve_style->gray_gc[5];
    int   dark=global.border_frame ? 4 : 5;

    FN_CHECK

    gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget), state, area, x,
                                       y, width, height);
    sanitize_size(window, &width, &height);

    switch(shadow_type)
    {
        case GTK_SHADOW_NONE:
            return;
        case GTK_SHADOW_ETCHED_IN:
        case GTK_SHADOW_IN:
            gc1 = qtcurve_style->gray_gc[dark];
            gc2 = qtcurve_style->gray_gc[0];
            break;
        case GTK_SHADOW_ETCHED_OUT:
        case GTK_SHADOW_OUT:
            gc1 = qtcurve_style->gray_gc[0];
            gc2 = qtcurve_style->gray_gc[dark];
            break;
    }

    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, area);
        gdk_gc_set_clip_rectangle(gc2, area);
        gdk_gc_set_clip_rectangle(outer, area);
    }
 
    if(global.border_frame) 
        switch(shadow_type)
        {
            case GTK_SHADOW_NONE:
            case GTK_SHADOW_IN:
            case GTK_SHADOW_OUT:
            case GTK_SHADOW_ETCHED_IN:
            case GTK_SHADOW_ETCHED_OUT:
                switch(gap_side)
                {
                    case GTK_POS_TOP:
                        if(gap_x > 0)
                        {
                            gdk_draw_line(window, gc1, x, y+1, x+gap_x+1, y+1);
                            gdk_draw_line(window, gc1, x+gap_x+1, y, x+gap_x+1, y);
                            gdk_draw_line(window, outer, x, y, x+gap_x, y);
                        }
                        else
                            gdk_draw_line(window, gc1, x+1, y, x+1, y+1);
                        if((width -(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc1, x+gap_x+gap_width-2, y+1, x+width-2, y+1);
                            gdk_draw_line(window, gc2, x+gap_x+gap_width-2, y, x+gap_x+gap_width-2, y);
                            gdk_draw_line(window, outer, x+gap_x+gap_width-1, y, x+width-1, y);
                        }
                        else
                            gdk_draw_line(window, gc2, x+width-2, y, x+width-2, y);
                        gdk_draw_line(window, gc1, x+1, y+2, x+1, y + height-1);
                        gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
                        gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-1);
                        gdk_draw_line(window, outer, x, y, x, y + height-1);
                        gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
                        gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case  GTK_POS_BOTTOM:
                        if(gap_x > 0)
                        {
                            gdk_draw_line(window, gc1, x+gap_x+1, y+height-1, x+gap_x+1, y+height-2);
                            gdk_draw_line(window, gc2, x+1, y+height-2, x+gap_x, y+height-2);
                            gdk_draw_line(window, outer, x, y+height-1, x+gap_x, y+height-1);
                        }
                        else
                            gdk_draw_line(window, gc1, x+1, y+height-1, x+1, y+height-2);
                        if((width -(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc2, x+gap_x+gap_width-2, y+height-2, x+width-1, y+height-2);
                            gdk_draw_line(window, gc2, x+gap_x+gap_width-2, y+height-1, x+gap_x+gap_width-2,
                                          y+height-1);
                            gdk_draw_line(window, outer, x+gap_x+gap_width-1, y+height-1, x+width-1, y+height-1);
                        }
                        else
                            gdk_draw_line(window, gc2, x+width-2, y+height-1, x+width-2, y+height-1);
                        gdk_draw_line(window, gc1, x+1, y+1, x+width-2, y+1);
                        gdk_draw_line(window, gc1, x+1, y+1, x+1, y+height-3);
                        gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-2);
                        gdk_draw_line(window, outer, x, y, x+width-1, y);
                        gdk_draw_line(window, outer, x, y, x, y+height-1);
                        gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case GTK_POS_LEFT:
                        if(gap_x>0)
                        {
                            gdk_draw_line(window, gc1, x+1, y+1, x+1, y+gap_x+1);
                            gdk_draw_line(window, gc1, x, y+gap_x+1, x, y+gap_x+1);
                            gdk_draw_line(window, outer, x, y, x, y+gap_x);
                        }
                        else
                            gdk_draw_line(window, gc1, x, y+1, x+1, y+1);
                        if((height-(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc1, x+1, y+gap_x+gap_width-2, x+1, y+height-1);
                            gdk_draw_line(window, gc2, x, y+gap_x+gap_width-2, x, y+gap_x+gap_width-2);
                            gdk_draw_line(window, outer, x, y+gap_x+gap_width-1, x, y+height-1);
                        }
                        else
                            gdk_draw_line(window, gc2, x, y+height-2, x, y+height-2);
                        gdk_draw_line(window, gc1, x+1, y+1, x+width-2, y+1);
                        gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-2);
                        gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
                        gdk_draw_line(window, outer, x, y, x+width-1, y);
                        gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
                        gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
                        break;
                    case GTK_POS_RIGHT:
                        if(gap_x>0)
                        {
                            gdk_draw_line(window, gc2, x+width-2, y, x+width-2, y+gap_x);
                            gdk_draw_line(window, gc1, x+width-2, y+gap_x+1, x+width-1, y+gap_x+1);
                            gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+gap_x);
                        }
                        else
                            gdk_draw_line(window, gc1, x+width-2, y+1, x+width, y+1);
                        if((height-(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc2, x+width-2, y+gap_x+gap_width-2, x+width, y+gap_x+gap_width-2);
                            gdk_draw_line(window, gc2, x+width-2, y+gap_x+gap_width-1, x+width-2, y+height-2);
                            gdk_draw_line(window, outer, x+width-1, y+gap_x+gap_width-1, x+width-1, y+height-1);
                        }
                        else
                            gdk_draw_line(window, gc2, x+width-1, y+height-2, x+width-1, y+height-2);
                        gdk_draw_line(window, gc1, x+1, y+1, x+width-3, y+1);
                        gdk_draw_line(window, gc1, x+1, y+1, x+1, y+height-2);
                        gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
                        gdk_draw_line(window, outer, x, y, x+width-1, y);
                        gdk_draw_line(window, outer, x, y, x, y+height-1);
                        gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
                        break;
                }
        }
    else
        switch(shadow_type)
        {
            case GTK_SHADOW_NONE:
            case GTK_SHADOW_IN:
            case GTK_SHADOW_OUT:
            case GTK_SHADOW_ETCHED_IN:
            case GTK_SHADOW_ETCHED_OUT:
                switch(gap_side)
                {
                    case GTK_POS_TOP:
                        if(gap_x > 0)
                            gdk_draw_line(window, gc1, x, y, x+gap_x, y);
                        if((width -(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc1, x+gap_x+gap_width, y, x+width-1, y);
                            gdk_draw_line(window, gc2, x+gap_x+gap_width-1, y, x+gap_x+gap_width-1, y);
                        }
                        gdk_draw_line(window, gc1, x, y, x, y + height-1);
                        gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
                        gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case  GTK_POS_BOTTOM:
                        if(gap_x > 0)
                            gdk_draw_line(window, gc2, x, y+height-1, x+gap_x, y+height-1);
                        if((width -(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc2, x+gap_x+gap_width, y+height-1, x+width-1, y+height-1);
                            gdk_draw_line(window, gc2, x+gap_x+gap_width-1, y+height-1, x+gap_x+gap_width-1,
                                          y+height-1);
                        }
                        gdk_draw_line(window, gc1, x, y, x+width-1, y);
                        gdk_draw_line(window, gc1, x, y, x, y+height-1);
                        gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case GTK_POS_LEFT:
                        if(gap_x>0)
                            gdk_draw_line(window, gc1, x, y, x, y+gap_x);
                        if((height-(gap_x + gap_width)) > 0)
                        {
                            gdk_draw_line(window, gc1, x, y+gap_x+gap_width, x, y+height-1);
                            gdk_draw_line(window, gc2, x, y+gap_x+gap_width-1, x, y+gap_x+gap_width-1);
                        }
                        gdk_draw_line(window, gc1, x, y, x+width-1, y);
                        gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
                        gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
                        break;
                    case GTK_POS_RIGHT:
                        if(gap_x>0)
                        {
                            gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+gap_x);
                            gdk_draw_line(window, gc1, x+width-1, y+gap_x, x+width-1, y+gap_x);
                        }
                        if((height-(gap_x + gap_width)) > 0)
                            gdk_draw_line(window, gc2, x+width-1, y+gap_x+gap_width-1, x+width-1, y+height-1);
                        gdk_draw_line(window, gc1, x, y, x+width-1, y);
                        gdk_draw_line(window, gc1, x, y, x, y+height-1);
                        gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
                        break;
                }
        }

    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, NULL);
        gdk_gc_set_clip_rectangle(gc2, NULL);
        gdk_gc_set_clip_rectangle(outer, NULL);
    }
}

static void fill_tab(GtkStyle *style, GdkWindow *window, GdkRectangle *area, GtkStateType state, GdkGC *fill,
                     GdkColor *col, int x, int y, int width, int height, gboolean horiz, gboolean increase)
{
    if(global.gradient_effect)
    {
        double s1=increase
                      ? GTK_STATE_NORMAL==state ? SHADE_TAB_SEL_LIGHT : SHADE_TAB_LIGHT
                      : GTK_STATE_NORMAL==state ? SHADE_BOTTOM_TAB_SEL_DARK : SHADE_BOTTOM_TAB_DARK,
               s2=increase
                      ? GTK_STATE_NORMAL==state ? SHADE_TAB_SEL_DARK : SHADE_TAB_DARK
                      : GTK_STATE_NORMAL==state ? SHADE_BOTTOM_TAB_SEL_LIGHT : SHADE_BOTTOM_TAB_LIGHT;

        draw_bevel_gradient(window, fill, style->colormap, area, NULL, x, y, width, height, col, s1, s2, horiz,
                            increase);
    }
    else
        gdk_draw_rectangle(window, fill, TRUE, x, y, width, height);
}

static void draw_extension(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                           GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                           gint height, GtkPositionType gap_side)
{
    QT_CURVE_STYLE_VAR
    GdkGC    *gc1 = NULL,
             *gc2 = NULL,
             *fill = GTK_STATE_NORMAL==state ? style->bg_gc[GTK_STATE_NORMAL] : qtcurve_style->gray_gc[2],
             *outer = qtcurve_style->gray_gc[5];
    int      dark=global.border_frame ? 4 : 5;
    GdkColor *col=  GTK_STATE_NORMAL==state ? &(style->bg[GTK_STATE_NORMAL]) : &(qtcurve_style->gray[2]);
    FN_CHECK

#if 0
printf("Draw extension %d %d %d %d %d %d %d %s  ", state, shadow_type, gap_side, x, y, width, height, detail ? detail :
"NULL");
display_widget(widget, 3);
#endif

    gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget), GTK_STATE_NORMAL, area, x, y, width,
height);
    sanitize_size(window, &width, &height);

    switch(shadow_type)
    {
        case GTK_SHADOW_NONE:
            return;
        case GTK_SHADOW_IN:
        case GTK_SHADOW_ETCHED_IN:
            gc1 = qtcurve_style->gray_gc[dark];
            gc2 = qtcurve_style->gray_gc[0];
            break;
        case GTK_SHADOW_OUT:
        case GTK_SHADOW_ETCHED_OUT:
            gc1 = qtcurve_style->gray_gc[0];
            gc2 = qtcurve_style->gray_gc[dark];
            break;
    }

    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, area);
        gdk_gc_set_clip_rectangle(gc2, area);
        gdk_gc_set_clip_rectangle(fill, area);
        gdk_gc_set_clip_rectangle(outer, area);
    }

    if(global.border_frame)
        switch(shadow_type)
        {
            case GTK_SHADOW_NONE:
            case GTK_SHADOW_IN:
            case GTK_SHADOW_OUT:
            case GTK_SHADOW_ETCHED_IN:
            case GTK_SHADOW_ETCHED_OUT:
                switch(gap_side)
                {
                    case GTK_POS_TOP:
                        fill_tab(style, window, area, state, fill, col, x, y, width, height-1, TRUE, FALSE);
                        gdk_draw_line(window, gc1, x+1, y , x+1, y+height-1);
                        gdk_draw_line(window, gc2, x+1, y+height-2, x+width-2, y+height-2);
                        gdk_draw_line(window, gc2, x+width-2, y, x+width-2, y+height-1);
                        gdk_draw_line(window, outer, x, y , x, y+height-2);
                        gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-2);
                        if(global.rounded)
                        {
                            gdk_draw_line(window, outer, x+1, y+height-1, x+width-2, y+height-1);
                            if(GTK_STATE_NORMAL==state)
                            {
                                gdk_draw_point(window, outer, x+1, y+height-2);
                                gdk_draw_point(window, outer, x+width-2, y+height-2);
                                gdk_draw_line(window, gc2, x, y+height-2, x+1, y+height-1);
                                gdk_draw_line(window, gc2, x+width-2, y+height-1, x+width-1, y+height-2);
                            }
                        }
                        else
                            gdk_draw_line(window, outer, x, y+height-1, x+width-1, y+height-1);
                        break;
                    case GTK_POS_BOTTOM:
                        fill_tab(style, window, area, state, fill, col, x, y+1, width, height-1, TRUE, TRUE);
                        gdk_draw_line(window, gc1, x+1, y+1 , x+width-2, y+1);
                        gdk_draw_line(window, gc1, x+1, y+1, x+1, y+height-1);
                        gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-1);
                        gdk_draw_line(window, outer, x, y+1, x, y+height-1);
                        gdk_draw_line(window, outer, x+width-1, y+1, x+width-1, y+height-1);
                        if(global.rounded)
                        {
                            gdk_draw_line(window, outer, x+1, y , x+width-2, y);
                 
                            if(GTK_STATE_NORMAL==state)
                            {
                                gdk_draw_point(window, outer, x+1, y+1);
                                gdk_draw_point(window, outer, x+width-2, y+1);
                                gdk_draw_line(window, gc2, x, y+1, x+1, y);
                                gdk_draw_line(window, gc2, x+width-2, y, x+width-1, y+1);
                            }
                        }
                        else
                            gdk_draw_line(window, outer, x, y , x+width-1, y);
                        break;
                    case GTK_POS_LEFT:
                        fill_tab(style, window, area, state, fill, col, x, y, width-1, height, FALSE, FALSE);
                        gdk_draw_line(window, gc1, x, y+1, x+width-1, y+1);
                        gdk_draw_line(window, gc2, x+width-2, y+1, x+width-2, y+height-2);
                        gdk_draw_line(window, gc2, x, y+height-2, x+width-1, y+height-2);
                        gdk_draw_line(window, outer, x, y, x+width-2, y);
                        gdk_draw_line(window, outer, x, y+height-1, x+width-2, y+height-1);
                        if(global.rounded)
                        {
                            gdk_draw_line(window, outer, x+width-1, y+1, x+width-1, y+height-2);
                            if(GTK_STATE_NORMAL==state)
                            {
                                gdk_draw_point(window, outer, x+width-2, y+1);
                                gdk_draw_point(window, outer, x+width-2, y+height-2);
                                gdk_draw_line(window, gc2, x+width-2, y, x+width-1, y+1);
                                gdk_draw_line(window, gc2, x+width-2, y+height-1, x+width-1, y+height-2);
                            }
                        }
                        else
                            gdk_draw_line(window, outer, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case GTK_POS_RIGHT:
                        fill_tab(style, window, area, state, fill, col, x+1, y, width-1, height, FALSE, TRUE);
                        gdk_draw_line(window, gc1, x+1, y+1, x+1, y+height-2);
                        gdk_draw_line(window, gc1, x, y+1, x+width-1, y+1);
                        gdk_draw_line(window, gc2, x, y+height-2, x+width-1, y+height-2);
                        gdk_draw_line(window, outer, x+1, y, x+width-1, y);
                        gdk_draw_line(window, outer, x+1, y+height-1, x+width-1, y+height-1);
                        if(global.rounded)
                        {
                            gdk_draw_line(window, outer, x, y+1, x, y+height-2);
                            if(GTK_STATE_NORMAL==state)
                            {
                                gdk_draw_point(window, outer, x+1, y+1);
                                gdk_draw_point(window, outer, x+1, y+height-2);
                                gdk_draw_line(window, gc2, x, y+1, x+1, y);
                                gdk_draw_line(window, gc2, x, y+height-2, x+1, y+height-1);
                            }
                        }
                        else
                            gdk_draw_line(window, outer, x, y, x, y+height-1);
                        break;
                }
        }
    else
    {
        switch(shadow_type)
        {
            case GTK_SHADOW_NONE:
            case GTK_SHADOW_IN:
            case GTK_SHADOW_OUT:
            case GTK_SHADOW_ETCHED_IN:
            case GTK_SHADOW_ETCHED_OUT:
                switch(gap_side)
                {
                    case GTK_POS_TOP:
                        fill_tab(style, window, area, state, fill, col, x, y, width, height-1, TRUE, FALSE);
                        gdk_draw_line(window, gc1, x, y , x, y+height-1);
                        gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
                        gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case GTK_POS_BOTTOM:
                        fill_tab(style, window, area, state, fill, col, x, y+1, width, height-1, TRUE, TRUE);
                        gdk_draw_line(window, gc1, x, y , x+width-1, y);
                        gdk_draw_line(window, gc1, x, y, x, y+height-1);
                        gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
                        break;
                    case GTK_POS_LEFT:
                        fill_tab(style, window, area, state, fill, col, x, y, width-1, height, FALSE, FALSE);
                        gdk_draw_line(window, gc1, x, y , x+width-1, y);
                        gdk_draw_line(window, gc2, x+width-1, y, x+width-1, y+height-1);
                        gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
                        break;
                    case GTK_POS_RIGHT:
                        fill_tab(style, window, area, state, fill, col, x+1, y, width-1, height, FALSE, TRUE);
                        gdk_draw_line(window, gc1, x, y , x, y+height-1);
                        gdk_draw_line(window, gc1, x, y, x+width-1, y);
                        gdk_draw_line(window, gc2, x, y+height-1, x+width-1, y+height-1);
                        break;
                }
        }
    }

    if(area)
    {
        gdk_gc_set_clip_rectangle(gc1, NULL);
        gdk_gc_set_clip_rectangle(gc2, NULL);
        gdk_gc_set_clip_rectangle(fill, NULL);
        gdk_gc_set_clip_rectangle(outer, NULL);
    }
}

static void draw_handle(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                        GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                        gint height, GtkOrientation orientation)
{
    QT_CURVE_STYLE_VAR
    FN_CHECK
  
    sanitize_size(window, &width, &height);
    gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget), state, area, x,
                                       y, width, height);

#if GTK_MAJOR_VERSION>1
/* CPD: From Industrial */
    if (DETAIL("dockitem") || WIDGET_TYPE_NAME ("PanelAppletFrame"))
        if(orientation == GTK_ORIENTATION_HORIZONTAL)
            orientation=GTK_ORIENTATION_VERTICAL;
        else
            orientation=GTK_ORIENTATION_HORIZONTAL;
#endif
    if(detail && (!strcmp(detail, "paned") || !strcmp(detail+1, "paned")))
        draw_box(style, window, GTK_STATE_NORMAL, GTK_SHADOW_OUT, area, widget, QTC_PANED, x, y, width, height);
    else if((DETAIL("handlebox") && widget && GTK_IS_HANDLE_BOX(widget)) || DETAIL("dockitem"))
    {
        draw_box(style, window, state, shadow_type, area, widget, "handlebox", x, y, width, height);
        draw_lines(window, x, y, width, height, height<width, 2, 4, qtcurve_style->gray_gc, area,
                   GROOVE_SUNKEN==global.handles);
    }
}

#define HIGHLIGHT_MOD 1.05

static void draw_flat_box(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                          GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                          gint height)
{
    if(global.radio_check_label_highlight && GTK_STATE_PRELIGHT==state &&
       (GTK_IS_RADIO_BUTTON(widget) || GTK_IS_CHECK_BUTTON(widget)))
    {
        GdkColor    highlight;
        GdkGCValues old_values;
        GdkGC       *gc=style->bg_gc[GTK_STATE_PRELIGHT];

        highlight.red=style->bg[GTK_STATE_PRELIGHT].red*HIGHLIGHT_MOD;
        highlight.green=style->bg[GTK_STATE_PRELIGHT].green*HIGHLIGHT_MOD;
        highlight.blue=style->bg[GTK_STATE_PRELIGHT].blue*HIGHLIGHT_MOD;
        highlight.pixel=0;

        gdk_gc_get_values(gc, &old_values);

        if(area)
            gdk_gc_set_clip_rectangle(gc, area);

        gdk_rgb_find_color(style->colormap, &highlight);
        gdk_gc_set_foreground(gc, &highlight);
        gdk_draw_rectangle(window, gc, TRUE, x, y, width, height);
        gdk_gc_set_foreground(gc, &old_values.foreground);

        if(area)
            gdk_gc_set_clip_rectangle(gc, NULL);
    }
    else
        parent_class->draw_flat_box(style, window, state, shadow_type, area, widget, detail, x, y, width,
                                    height);
}


static void draw_box(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                     GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                     gint height)
{
    QT_CURVE_STYLE_VAR
    gboolean custom_c = FALSE,
             pbar=DETAIL("bar") && GTK_IS_PROGRESS_BAR(widget),
             qtc_paned=IS_QTC_PANED;
    GdkGC    *new_gcs[7],
             **btn_gcs=NULL;
    GdkColor new_cols[7],
             *btn_cols;
    int      bgnd=0;
#if 0
printf("Draw box %d %d %s  ", state, shadow_type, detail ? detail : "NULL");
display_widget(widget, 3);
#endif

    if(use_button_color(detail))
    {
#if GTK_MAJOR_VERSION>1
        int btn_down=GTK_IS_BUTTON(widget) && GTK_BUTTON(widget)->depressed;

        bgnd=state==GTK_STATE_PRELIGHT ? btn_down ? 2 : 1 :
             state==GTK_STATE_ACTIVE || btn_down ? 3 : 6;
#else
        bgnd=state==GTK_STATE_PRELIGHT ? 1 :
             state==GTK_STATE_ACTIVE ? 3 : 6;
#endif

        if(QT_CUSTOM_COLOR_BUTTON(style))
        {
            custom_c=TRUE;
            shade_colors(&(style->bg[state]), new_cols);
            GEN_GCS(style, new_cols, new_gcs);
            btn_gcs=new_gcs;
            btn_cols=new_cols;
        }
        else
        {
            btn_gcs=qtcurve_style->button_gc;
            btn_cols=qtcurve_style->button;
        }
    }
  
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if((-1==width) &&(-1==height))
        gdk_window_get_size(window, &width, &height);
    else if(-1==width)
        gdk_window_get_size(window, &width, NULL);
    else if(-1==height)
        gdk_window_get_size(window, NULL, &height);

    if(detail && (0==strcmp(detail, "spinbutton_up") || 0==strcmp(detail, "spinbutton_down")))
    {
#if GTK_MAJOR_VERSION>1
        x--;
        y--;
        width+=2;
        height+=2;
#else
        height++;
        if(0==strcmp(detail, "spinbutton_down"))
        {
            y--;
            height++;
        }
#endif

        draw_bgnd(window, btn_gcs[bgnd], widget, area, x+1, y+1, width-2, height-2);

        if(global.gradient_effect)
        {
            int      sunken=2==bgnd || 3==bgnd,
                     offset=global.border_button && !sunken ? 2 : 1;

            draw_bevel_gradient(window, btn_gcs[bgnd], style->colormap, area, NULL, x+offset, y+offset,
                                width-(2*offset), height-(2*offset), &btn_cols[bgnd],
                                sunken ? SHADE_BEVEL_GRAD_SEL_LIGHT : SHADE_BEVEL_GRAD_LIGHT,
                                sunken ? SHADE_BEVEL_GRAD_SEL_DARK : SHADE_BEVEL_GRAD_DARK, TRUE, !sunken);

        }

        gtk_paint_shadow(style, window, state, shadow_type, area, widget, detail, x+1, y+1, width-2, height-2);
    }
    else if(DETAIL("spinbutton"))
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
                                           GTK_STATE_INSENSITIVE==state 
                                               ? GTK_STATE_INSENSITIVE 
                                               : GTK_STATE_NORMAL, 
                                           area, x, y, width, height);
    else if(detail &&( 0==strcmp(detail, "button") ||
                       0==strcmp(detail, "togglebutton") ||
                       /*0==strcmp(detail, "spinbutton") || */
                       0==strcmp(detail, "slider") ||
                       0==strcmp(detail, "hscrollbar") ||
                       0==strcmp(detail, "vscrollbar")  ||
                       0==strcmp(detail, "hscale") ||
                       0==strcmp(detail, "vscale") ||
                       qtc_paned))
    {
        if(is_button_on_toolbar(widget))
        {
            y+=1;
            height-=2;
        }

        draw_bgnd(window, btn_gcs[bgnd], widget, area, x, y, width, height);

        if(!qtc_paned && global.gradient_effect)
        {
            int      sunken=2==bgnd || 3==bgnd,
                     offset=global.border_button && !sunken ? 2 : 1;
            gboolean horiz=(0==strcmp(detail, "slider") && width<height) || 0==strcmp(detail, "vscrollbar") || 0==strcmp(detail,
"vscale")
                               ? FALSE
                               : TRUE;

            draw_bevel_gradient(window, btn_gcs[bgnd], style->colormap, area, NULL, x+offset, y+offset,
                                width-(2*offset), height-(2*offset), &btn_cols[bgnd], 
                                sunken ? SHADE_BEVEL_GRAD_SEL_LIGHT : SHADE_BEVEL_GRAD_LIGHT,
                                sunken ? SHADE_BEVEL_GRAD_SEL_DARK : SHADE_BEVEL_GRAD_DARK, horiz, !sunken);
        }

        if(IND_CORNER==global.def_btn_indicator && 0==strcmp(detail, "button") && GTK_WIDGET_HAS_DEFAULT(widget))
        {
            int      sunken=2==bgnd || 3==bgnd,
                     offset=sunken ? 4 : 3;
            GdkGC    *gc=btn_gcs[GTK_STATE_ACTIVE==state ? 0 : global.border_button ? 4 : 5];
            GdkPoint points[3] = { { x+offset, y+offset }, { x+offset+6, y+offset}, { x+offset, y+offset+6}};

            if(area)
                gdk_gc_set_clip_rectangle(gc, area);

            gdk_draw_polygon(window, gc, TRUE, points, 3);

            if(area)
                gdk_gc_set_clip_rectangle(gc, NULL);
        }

        gtk_paint_shadow(style, window, state, shadow_type, area, widget, qtc_paned ? QTC_PANED : detail, x, y,
                         width, height);
    }
    else if(detail && (0==strcmp(detail, "buttondefault") || 0==strcmp(detail, "togglebuttondefault")))
    {
        if(IND_BORDER==global.def_btn_indicator)
        {
            GdkGC *def_gc=/*global.border_button || */global.rounded
                              ? btn_gcs[4]
                              : style->text_gc[state];
#if GTK_MAJOR_VERSION>1
#define XM x
#define YM y
#define HM height
#define WM width
#else
            gint xmod=QT_STYLE->xthickness+2,
                 ymod=QT_STYLE->ythickness+2,
                 hmod=2*xmod,
                 wmod=2*ymod;
#define XM (x+xmod)
#define YM (y+ymod)
#define HM (height-hmod)
#define WM (width-wmod)
#endif
            if(area)
                gdk_gc_set_clip_rectangle(def_gc, area);

            if(global.rounded)
            {
                int offset=width>=QTC_MIN_BTN_SIZE && height>=QTC_MIN_BTN_SIZE ? 4 : 3;

                gdk_draw_line(window, def_gc, XM+offset, YM, XM+WM-(1+offset), YM);
                gdk_draw_line(window, def_gc, XM+offset, YM+HM-1, XM+WM-(1+offset), YM+HM-1);
                gdk_draw_line(window, def_gc, XM, YM+offset, XM, YM+HM-(1+offset));                                   
                gdk_draw_line(window, def_gc, XM+WM-1, YM+offset, XM+WM-1, YM+HM-(1+offset));
#if GTK_MAJOR_VERSION==1
                /* GTK1 shrinks the button when pressed ! */
                offset--;
                gdk_draw_line(window, def_gc, XM+offset, YM+1, XM+WM-(1+offset), YM+1);
                gdk_draw_line(window, def_gc, XM+offset, YM+HM-2, XM+WM-(1+offset), YM+HM-2);
                gdk_draw_line(window, def_gc, XM+1, YM+offset, XM+1, YM+HM-(1+offset));
                gdk_draw_line(window, def_gc, XM+WM-2, YM+offset, XM+WM-2, YM+HM-(1+offset));
#endif
            }
            else
                gdk_draw_rectangle(window, def_gc, TRUE, XM, YM, WM, HM);
#undef XM
#undef YM
#undef HM
#undef WM
            if(area)
                gdk_gc_set_clip_rectangle(def_gc, NULL);
        }
    }
    else if(widget && DETAIL("trough")) /* Scrollbars */
    {
        gboolean pbar=GTK_IS_PROGRESS_BAR(widget);
        GdkGC *bgnd=pbar ? qtcurve_style->gray_gc[6] : qtcurve_style->gray_gc[2],
              *gcl =pbar ? qtcurve_style->gray_gc[0] : qtcurve_style->gray_gc[4],
              *gcd =/*pbar ? qtcurve_style->gray_gc[5] : */qtcurve_style->gray_gc[5];

        if(area)
        {
            gdk_gc_set_clip_rectangle(bgnd, area);
            gdk_gc_set_clip_rectangle(gcl, area);
            gdk_gc_set_clip_rectangle(gcd, area);
        }

        if(GTK_IS_HSCALE(widget) || GTK_IS_VSCALE(widget))
        {
            gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget),
                                               GTK_STATE_INSENSITIVE==state
                                                   ? GTK_STATE_INSENSITIVE 
                                                   : GTK_STATE_NORMAL, 
                                               area, x, y, width, height);
            if GTK_IS_HSCALE(widget)
            {
                y +=(height - SCALE_SIZE)>>1;
                height = SCALE_SIZE;
            }
            else
            {
                x +=(width - SCALE_SIZE)>>1;
                width = SCALE_SIZE;
            }

            gdk_draw_rectangle(window, bgnd, TRUE, x, y, width, height);
            gdk_draw_rectangle(window, gcd, FALSE, x, y, width - 1, height - 1);
            gdk_draw_line(window, gcl, x + 1, y + 1, x + width - 2, y + 1);
            gdk_draw_line(window, gcl, x + 1, y + 1, x + 1, y + height - 2);
        }
        else
        {
            gdk_draw_rectangle(window, bgnd, TRUE, x, y, width, height);

            if(GTK_IS_PROGRESS_BAR(widget))
            {
                gdk_draw_line(window, gcd, x,         y,          x+width-2, y);
                gdk_draw_line(window, gcd, x,         y,          x,         y+height-2);
                gdk_draw_line(window, gcl, x,         y+height-1, x+width-1, y+height-1);
                gdk_draw_line(window, gcl, x+width-1, y+height-1, x+width-1, y);
            }
            else if(global.border_button)
                gdk_draw_rectangle(window, gcd, FALSE, x, y, width - 1, height - 1);
        }

        if(area)
        {
            gdk_gc_set_clip_rectangle(bgnd, NULL);
            gdk_gc_set_clip_rectangle(gcl, NULL);
            gdk_gc_set_clip_rectangle(gcd, NULL);
        }
    }
    else if(widget && ( (detail && ( 0==strcmp(detail, "menubar") || 0==strcmp(detail,"toolbar") ||
                                     0==strcmp(detail,"dockitem") || 0==strcmp(detail,"dockitem_bin") || 
                                     0==strcmp(detail, "handlebox") || 0==strcmp(detail,"handlebox_bin") ) )
                        || WIDGET_TYPE_NAME("PanelAppletFrame")) && shadow_type != GTK_SHADOW_NONE)
    {
        /* Toolbars and menus */
        gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget), state, area,
                                           x, y, width, height);

        if(TB_NONE!=global.toolbar_borders) 
        {
            gboolean top=FALSE,
                     bottom=FALSE,
                     left=FALSE,
                     right=FALSE;
            int      border=TB_DARK==global.toolbar_borders ? 5 : 3;

            if(area)
            {
                gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[0], area);
                gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[border], area);
            }

            if(0==strcmp(detail, "menubar"))
                top=bottom=TRUE;
            else if(0==strcmp(detail,"toolbar") && GTK_IS_TOOLBAR(widget))
            {
#if GTK_MAJOR_VERSION>1
                if(GTK_ORIENTATION_HORIZONTAL==gtk_toolbar_get_orientation(GTK_TOOLBAR(widget)))
#else
                if(GTK_ORIENTATION_HORIZONTAL== GTK_PROGRESS_BAR(widget)->orientation)
#endif
                    top=bottom=TRUE;
                else   /* MUST have a handle */
                    left=right=bottom=TRUE;
            }
            else if(0==strcmp(detail,"dockitem_bin"))  /* CPD: bit risky - what if only 1 item ??? */
            {
                if(width<height)
                    left=right=top=TRUE;
                else
                    top=bottom=TRUE;
            }
            else /* handle */
                if(width<height) /* on horiz toolbar */
                    top=bottom=TRUE;
                else
                    left=right=top=TRUE;

            if(top)
                gdk_draw_line(window, qtcurve_style->gray_gc[0], x, y, x + width-1, y);
            if(left)
                gdk_draw_line(window, qtcurve_style->gray_gc[0], x, y, x, y+height-1);
            if(bottom)
                gdk_draw_line(window, qtcurve_style->gray_gc[border], x, y + height - 1, x + width-1, y + height - 1);
            if(right)
                gdk_draw_line(window, qtcurve_style->gray_gc[border], x+width-1, y, x + width-1, y + height - 1);
      
            if(area)
            {
                gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[0], NULL);
                gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[border], NULL);
            }
        }
    }
    else if(widget &&(DETAIL("menuitem") || pbar))
    {
        if(global.gradient_effect_pm)
        {
            int offset=global.border_frame ? 2 : 1;

            if(area)
            {
                gdk_gc_set_clip_rectangle(style->base_gc[state], area);
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_BASE], area);
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_LIGHT], area);
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_DARK], area);
            }

            if(pbar)
            {
                GtkProgressBarOrientation
#if GTK_MAJOR_VERSION>1
                  orientation = gtk_progress_bar_get_orientation(GTK_PROGRESS_BAR(widget));
#else
                  orientation = GTK_PROGRESS_BAR(widget)->orientation;
#endif
                gboolean horiz=GTK_PROGRESS_LEFT_TO_RIGHT==orientation || GTK_PROGRESS_RIGHT_TO_LEFT==orientation;

                if((horiz && width > 2) || (!horiz && height > 2))
                    draw_gradient(window, style->base_gc[state], style->colormap, area, NULL, x+offset, y+offset,
                                  width-(2*offset), height-(2*offset), &qtcurve_style->gradient[GRADIENT_TOP], 
                                  &qtcurve_style->gradient[GRADIENT_BOTTOM], horiz, TRUE);
            }
            else
                draw_gradient(window, style->base_gc[state], style->colormap, area, NULL, x+offset, y+offset,
                              width-(2*offset), height-(2*offset), &qtcurve_style->gradient[GRADIENT_TOP],
                              &qtcurve_style->gradient[GRADIENT_BOTTOM], TRUE, TRUE);


            if(global.border_frame)
                gdk_draw_rectangle(window, qtcurve_style->gradient_gc[GRADIENT_BASE], FALSE, x, y, width-1, height-1);
            else
            {
                x--; y--; width+=2; height+=2;
            }
            gdk_draw_line(window, qtcurve_style->gradient_gc[GRADIENT_LIGHT], x+1, y+1, x+1, y+height-2);
            gdk_draw_line(window, qtcurve_style->gradient_gc[GRADIENT_LIGHT], x+1, y+1, x+width-2, y+1);
            gdk_draw_line(window, qtcurve_style->gradient_gc[GRADIENT_DARK], x+1, y+height-2, x+width-2, y+height-2);
            gdk_draw_line(window, qtcurve_style->gradient_gc[GRADIENT_DARK], x+width-2, y+height-2, x+width-2, y+1);

            if(area)
            {
                gdk_gc_set_clip_rectangle(style->base_gc[state], NULL);
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_BASE], NULL);
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_LIGHT], NULL);
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_DARK], NULL);
            }
        }
        else
        {
            if(area)
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_BASE], area);
            gdk_draw_rectangle(window, qtcurve_style->gradient_gc[GRADIENT_BASE], TRUE, x, y, width, height);
            if(area)
                gdk_gc_set_clip_rectangle(qtcurve_style->gradient_gc[GRADIENT_BASE], NULL);
        }
    }
    else if(widget && DETAIL("optionmenu"))
    {
        GtkRequisition indicator_size;
        GtkBorder      indicator_spacing;

        if(!style->bg_pixmap[state]
#if GTK_MAJOR_VERSION>1
                || GDK_IS_PIXMAP(window)
#endif
         )
        {
            draw_bgnd(window, btn_gcs[bgnd], widget, area, x, y, width, height);
            if(global.gradient_effect)
            {
                int sunken=2==bgnd || 3==bgnd,
                    offset=global.border_button && !sunken ? 2 : 1;

                draw_bevel_gradient(window, btn_gcs[bgnd], style->colormap, area, NULL, x+offset, y+offset,
                                    width-(2*offset), height-(2*offset), &btn_cols[bgnd],
                                    sunken ? SHADE_BEVEL_GRAD_SEL_LIGHT : SHADE_BEVEL_GRAD_LIGHT,
                                    sunken ? SHADE_BEVEL_GRAD_SEL_DARK : SHADE_BEVEL_GRAD_DARK, TRUE, !sunken);
            }
        }
        else
            gtk_style_apply_default_background(style, window, widget && !GTK_WIDGET_NO_WINDOW(widget), 
                                               state, area, x, y, width, height);

        gtk_paint_shadow(style, window, state, shadow_type, area, widget, detail, x, y, width, height);
        option_menu_get_props(widget, &indicator_size, &indicator_spacing);

        if(area)
        {
            gdk_gc_set_clip_rectangle(btn_gcs[0], area);
            gdk_gc_set_clip_rectangle(btn_gcs[4], area);
        }

#if (GTK_MAJOR_VERSION>1) && (GTK_MINOR_VERSION<2)
        y++;
        height-=2;
#endif
        y+=3;
        height-=6;

        gdk_draw_line(window, btn_gcs[global.border_button ? 4 : 5],
                      x + width -(indicator_size.width + indicator_spacing.left + indicator_spacing.right) - 
                          QT_STYLE->xthickness,
                      y + QT_STYLE->ythickness,
                      x + width -(indicator_size.width + indicator_spacing.left + indicator_spacing.right) - 
                          QT_STYLE->xthickness,
                      y + height - QT_STYLE->ythickness-1);
    
        gdk_draw_line(window, btn_gcs[0],
               x + width -(indicator_size.width + indicator_spacing.left + indicator_spacing.right) - 
                   QT_STYLE->xthickness + 1,
               y + QT_STYLE->ythickness,
               x + width -(indicator_size.width + indicator_spacing.left + indicator_spacing.right) - 
                   QT_STYLE->xthickness + 1,
               y + height - QT_STYLE->ythickness - 1);
    
             
        if(area)
        {
            gdk_gc_set_clip_rectangle(btn_gcs[0], NULL);
            gdk_gc_set_clip_rectangle(btn_gcs[4], NULL);
        }
    }
    else
        parent_class->draw_box(style, window, state, shadow_type, area, widget, detail, x, y, width, height);

#if GTK_MAJOR_VERSION==1
    if(DETAIL("slider"))
    {
        const char     *slider_detail = detail;
        GtkOrientation orientation = GTK_ORIENTATION_VERTICAL;
      
        if(widget)
        {
            if(GTK_IS_HSCALE(widget))
            {
                slider_detail = "hscale";
                orientation = GTK_ORIENTATION_HORIZONTAL;
            }
            else if(GTK_IS_VSCALE(widget))
            {
                slider_detail = "vscale";
                orientation = GTK_ORIENTATION_VERTICAL;
            }
            else if(GTK_IS_HSCROLLBAR(widget))
                orientation = GTK_ORIENTATION_HORIZONTAL;
        }

        real_draw_slider(style, window, state, shadow_type, area, widget, slider_detail, x, y, width, height, 
                         orientation);
    }
#endif
    if(detail &&(!strcmp(detail, "paned") || !strcmp(detail+1, "paned")))
    {
        GtkOrientation orientation = GTK_ORIENTATION_HORIZONTAL;

        if(*detail == 'h')
            orientation = GTK_ORIENTATION_VERTICAL;

        draw_handle(style, window, state, shadow_type, area, widget, detail, x, y, width, height, orientation);
    }
    if(custom_c)
        RELEASE_GCS(new_gcs);

#if GTK_MAJOR_VERSION==1
        /* CPD: HACK - force trough to be drawn */
        if(widget && GTK_IS_RANGE(widget))
        {
            static int called=0;

            if(!called)  /* Prevent recursion */
            {
                called=1;
                gtk_range_draw_trough(widget);
                called=0;
            }
        }
#endif
}

#if GTK_MAJOR_VERSION>1
static void draw_slider
#else
static void real_draw_slider
#endif
               (GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type, 
                GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width, gint height,
                GtkOrientation orientation)
{
    QT_CURVE_STYLE_VAR
    gboolean custom_c = FALSE,
             horiz_scale = DETAIL("hscale");
    GdkGC    *new_gcs[7],
             **btn_gcs;

    if(use_button_color(detail))
        if(QT_CUSTOM_COLOR_BUTTON(style))
        {
            GdkColor new_colors[7];
            custom_c=TRUE;
            shade_colors(&(style->bg[state]), new_colors);
            GEN_GCS(style, new_colors, new_gcs);
            btn_gcs=new_gcs;
        }
        else
            btn_gcs=qtcurve_style->button_gc;

    FN_CHECK
    sanitize_size(window, &width, &height);

#if GTK_MAJOR_VERSION>1
    gtk_paint_box(style, window, state, shadow_type, area, widget, detail, x, y, width, height);
#endif

   /* Orientation is always vertical with Mozilla, why? Anyway this hack should be OK - as we only draw dashes when 
      slider is larger than 23 pixels... */
    orientation=width<height ? GTK_ORIENTATION_VERTICAL : GTK_ORIENTATION_HORIZONTAL;
    if(detail &&(0==strcmp("slider", detail) || horiz_scale || 0==strcmp("vscale", detail)) &&
      GROOVE_NONE!=global.slider_thumbs &&
      (((horiz_scale || GTK_ORIENTATION_HORIZONTAL==orientation) && width>=20) || height>=20))
        draw_lines(window, x, y, width, height, !(horiz_scale || GTK_ORIENTATION_HORIZONTAL==orientation), 4, 3, 
        btn_gcs, area, GROOVE_SUNKEN==global.slider_thumbs);

    if(custom_c)
        RELEASE_GCS(new_gcs);
}

#if GTK_MAJOR_VERSION==1
static void draw_slider(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type,
                        GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                        gint height, GtkOrientation orientation)
{
    QT_CURVE_STYLE_VAR
    FN_CHECK
    sanitize_size(window, &width, &height);
    gtk_paint_box(style, window, state, shadow_type, area, widget, detail, x, y, width, height);
    real_draw_slider(style, window, state, shadow_type, area, widget, detail, x, y, width, height, orientation);
}
#endif

static void draw_shadow_gap(GtkStyle *style, GdkWindow *window, GtkStateType state, GtkShadowType shadow_type, 
                           GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width,
                           gint height, GtkPositionType gap_side, gint gap_x, gint gap_width)
{
    QT_CURVE_STYLE_VAR
    GdkGC *outer_gc = NULL,
          *gc1 = NULL,
          *gc2 = NULL,
          *gc3 = NULL,
          *gc4 = NULL;
 
    FN_CHECK
    sanitize_size(window, &width, &height);
    outer_gc = qtcurve_style->gray_gc[5];
  
    switch(shadow_type)
    {
        case GTK_SHADOW_NONE:
            return;
        case GTK_SHADOW_IN:
            gc1 = qtcurve_style->gray_gc[5];
            gc2 = qtcurve_style->gray_gc[0];
            gc3 = qtcurve_style->gray_gc[0];
            gc4 = qtcurve_style->gray_gc[5];
            break;
        case GTK_SHADOW_ETCHED_IN:
            gc1 = qtcurve_style->gray_gc[5];
            gc2 = qtcurve_style->gray_gc[0];
            gc3 = qtcurve_style->gray_gc[5];
            gc4 = qtcurve_style->gray_gc[0];
            break;
        case GTK_SHADOW_OUT:
            gc1 = qtcurve_style->gray_gc[5];
            gc2 = qtcurve_style->gray_gc[0];
            gc3 = qtcurve_style->gray_gc[0];
            gc4 = qtcurve_style->gray_gc[5];
            break;
        case GTK_SHADOW_ETCHED_OUT:
            gc1 = qtcurve_style->gray_gc[0];
            gc2 = qtcurve_style->gray_gc[5];
            gc3 = qtcurve_style->gray_gc[0];
            gc4 = qtcurve_style->gray_gc[5];
            break;
    }

    if(area)
    {
        gdk_gc_set_clip_rectangle(outer_gc, area);
        gdk_gc_set_clip_rectangle(gc1, area);
        gdk_gc_set_clip_rectangle(gc2, area);
        gdk_gc_set_clip_rectangle(gc3, area);
        gdk_gc_set_clip_rectangle(gc4, area);
    }
  
    switch(shadow_type)
    {
        case GTK_SHADOW_NONE:
        case GTK_SHADOW_IN:
        case GTK_SHADOW_OUT:
        case GTK_SHADOW_ETCHED_IN:
        case GTK_SHADOW_ETCHED_OUT:
            switch(gap_side)
            {
                case GTK_POS_TOP:
                    gdk_draw_line(window, gc1, x, y, x, y + height - 1);
                    gdk_draw_line(window, gc2, x + 1, y, x + 1, y + height - 2);
                    gdk_draw_line(window, gc3, x + 1, y + height - 2, x + width - 2, y + height - 2);
                    gdk_draw_line(window, gc3, x + width - 2, y, x + width - 2, y + height - 2);
                    gdk_draw_line(window, gc4, x, y + height - 1, x + width - 1, y + height - 1);
                    gdk_draw_line(window, gc4, x + width - 1, y, x + width - 1, y + height - 1);
                    if(gap_x > 0)
                    {
                        gdk_draw_line(window, gc1, x, y, x + gap_x - 1, y);
                        gdk_draw_line(window, gc2, x + 1, y + 1, x + gap_x - 1, y + 1);
                    }
                    if((width -(gap_x + gap_width)) > 0)
                    {
                        gdk_draw_line(window, gc1, x + gap_x + gap_width, y, x + width - 2, y);
                        gdk_draw_line(window, gc2, x + gap_x + gap_width, y + 1, x + width - 3, y + 1);
                    }
                    break;
                case GTK_POS_BOTTOM:
                    gdk_draw_line(window, gc1, x, y, x + width - 1, y);
                    gdk_draw_line(window, gc1, x, y, x, y + height - 1);
                    gdk_draw_line(window, gc2, x + 1, y + 1, x + width - 2, y + 1);
                    gdk_draw_line(window, gc2, x + 1, y + 1, x + 1, y + height - 1);
                    gdk_draw_line(window, gc3, x + width - 2, y + 1, x + width - 2, y + height - 1);
                    gdk_draw_line(window, gc4, x + width - 1, y, x + width - 1, y + height - 1);
                    if(gap_x > 0)
                    {
                        gdk_draw_line(window, gc4, x, y + height - 1, x + gap_x - 1, y + height - 1);
                        gdk_draw_line(window, gc3, x + 1, y + height - 2, x + gap_x - 1, y + height - 2);
                    }
                    if((width -(gap_x + gap_width)) > 0)
                    {
                        gdk_draw_line(window, gc4, x + gap_x + gap_width, y + height - 1, x + width - 2, 
                                      y + height - 1);
                        gdk_draw_line(window, gc3, x + gap_x + gap_width, y + height - 2, x + width - 2, 
                                      y + height - 2);
                    }
                    break;
                case GTK_POS_LEFT:
                    gdk_draw_line(window, gc1, x, y, x + width - 1, y);
                    gdk_draw_line(window, gc2, x, y + 1, x + width - 2, y + 1);
                    gdk_draw_line(window, gc3, x, y + height - 2, x + width - 2, y + height - 2);
                    gdk_draw_line(window, gc3, x + width - 2, y + 1, x + width - 2, y + height - 2);
                    gdk_draw_line(window, gc4, x, y + height - 1, x + width - 1, y + height - 1);
                    gdk_draw_line(window, gc4, x + width - 1, y, x + width - 1, y + height - 1);
                    if(gap_x > 0)
                    {
                        gdk_draw_line(window, gc1, x, y, x, y + gap_x - 1);
                        gdk_draw_line(window, gc2, x + 1, y + 1, x + 1, y + gap_x - 1);
                    }
                    if((width -(gap_x + gap_width)) > 0)
                    {
                        gdk_draw_line(window, gc1, x, y + gap_x + gap_width, x, y + height - 2);
                        gdk_draw_line(window, gc2, x + 1, y + gap_x + gap_width, x + 1, y + height - 2);
                    }
                    break;
                case GTK_POS_RIGHT:
                    gdk_draw_line(window, gc1, x, y, x + width - 1, y);
                    gdk_draw_line(window, gc1, x, y, x, y + height - 1);
                    gdk_draw_line(window, gc2, x + 1, y + 1, x + width - 1, y + 1);
                    gdk_draw_line(window, gc2, x + 1, y + 1, x + 1, y + height - 2);
                    gdk_draw_line(window, gc3, x + 1, y + height - 2, x + width - 1, y + height - 2);
                    gdk_draw_line(window, gc4, x, y + height - 1, x + width - 1, y + height - 1);
                    if(gap_x > 0)
                    {
                        gdk_draw_line(window, gc4, x + width - 1, y, x + width - 1, y + gap_x - 1);
                        gdk_draw_line(window, gc3, x + width - 2, y + 1, x + width - 2, y + gap_x - 1);
                    }
                    if((width -(gap_x + gap_width)) > 0)
                    {
                        gdk_draw_line(window, gc4, x + width - 1, y + gap_x + gap_width, x + width - 1, 
                                      y + height - 2);
                        gdk_draw_line(window, gc3, x + width - 2, y + gap_x + gap_width, x + width - 2, 
                                      y + height - 2);
                    }
                    break;
            }
     }
  
    if(area)
    {
        gdk_gc_set_clip_rectangle(outer_gc, NULL);
        gdk_gc_set_clip_rectangle(gc1, NULL);
        gdk_gc_set_clip_rectangle(gc2, NULL);
        gdk_gc_set_clip_rectangle(gc3, NULL);
        gdk_gc_set_clip_rectangle(gc4, NULL);
    }
}

static void draw_hline(GtkStyle *style, GdkWindow *window, GtkStateType state, GdkRectangle *area, 
                       GtkWidget *widget, PCONST gchar *detail, gint x1, gint x2, gint y)
{
    QT_CURVE_STYLE_VAR
    FN_CHECK
  
    if(area)
        gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[5], area);
  
    if(DETAIL("label"))
    {
        if(state == GTK_STATE_INSENSITIVE)
            gdk_draw_line(window, qtcurve_style->gray_gc[0], x1 + 1, y + 1, x2 + 1, y + 1);   
        gdk_draw_line(window, style->text_gc[state], x1, y, x2, y);     
    }
    else
    {
        gdk_draw_line(window, qtcurve_style->gray_gc[5], x1, y, x2, y);
        if(DETAIL("menuitem"))
            gdk_draw_line(window, qtcurve_style->gray_gc[0], x1, y+1, x2, y+1);
    }
  
    if(area)
        gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[5], NULL);
}

static void draw_vline(GtkStyle *style, GdkWindow *window, GtkStateType state, GdkRectangle *area, 
                       GtkWidget *widget, PCONST gchar *detail, gint y1, gint y2, gint x)
{
    QT_CURVE_STYLE_VAR
    FN_CHECK
  
    if(area)
        gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[5], area);
  
    gdk_draw_line(window, qtcurve_style->gray_gc[5], x, y1, x, y2 - 1);
 
    if(DETAIL("toolbar"))
        gdk_draw_line(window, qtcurve_style->gray_gc[0], x+1, y1, x+1, y2 - 1);

    if(area)
        gdk_gc_set_clip_rectangle(qtcurve_style->gray_gc[5], NULL);
}

static void draw_focus(GtkStyle *style, GdkWindow *window, 
#if GTK_MAJOR_VERSION>1
                       GtkStateType state,
#endif
                       GdkRectangle *area, GtkWidget *widget, PCONST gchar *detail, gint x, gint y, gint width, 
                       gint height)
{
    QT_CURVE_STYLE_VAR
    GdkPoint points[5];
    GdkGC    *gc= qtcurve_style->gray_gc[5];
    gboolean free_dash_list = FALSE;
    gint     line_width = 1,
             hl=0;
    char     *dash_list =(char *)"\1\1";
    gint     dash_len;

    if(GTK_IS_EDITABLE(widget))
        return;

#if GTK_MAJOR_VERSION>1
    if(widget)
    {
        gtk_widget_style_get(widget, "focus-line-width", &line_width, "focus-line-pattern",(gchar *)&dash_list, NULL);
        free_dash_list = TRUE;
    }
#else
    if(GTK_IS_OPTION_MENU(widget) || GTK_IS_BUTTON(widget) || GTK_IS_RADIO_BUTTON(widget) || 
       GTK_IS_CHECK_BUTTON(widget))
    {
        if(GTK_IS_OPTION_MENU(widget))
        {
            x+=4;
            y+=4;
            width-=26;
            height-=7;
        }
        else if(!(detail && 0==strcmp(detail, "togglebutton")) &&(GTK_IS_RADIO_BUTTON(widget) || 
                GTK_IS_CHECK_BUTTON(widget)))
        {
            x+=18;
            width-=18;
        }
        else
        {
            x+=4;
            y+=4;
            width-=7;
            height-=7;
        }
    }
#endif

    sanitize_size(window, &width, &height);
 
    if(area)
        gdk_gc_set_clip_rectangle(gc, area);

    gdk_gc_set_line_attributes(gc, line_width, dash_list[0] ? GDK_LINE_ON_OFF_DASH : GDK_LINE_SOLID, GDK_CAP_BUTT, 
                               GDK_JOIN_MITER);

    if(detail && !strcmp(detail, "add-mode"))
    {
        if(free_dash_list)
            g_free(dash_list);
      
        dash_list =(char *)"\4\4";
        free_dash_list = FALSE;
    }

    hl=line_width>>1;
    points[0].x = x + hl;
    points[0].y = y + hl;
    points[1].x = x + width - line_width + hl;
    points[1].y = y + hl;
    points[2].x = x + width - line_width + hl;
    points[2].y = y + height - line_width + hl;
    points[3].x = x + hl;
    points[3].y = y + height - line_width + hl;
    points[4] = points[0];
    
    if(!dash_list[0])
        gdk_draw_lines(window, gc, points, 5);
    else
    {
        /* We go through all the pain below because the X rasterization rules don't really work right for dashed 
        * lines if you want continuity in segments that go between top/right and left/bottom. For instance, a top 
        * left corner with a 1-1 dash is drawn as:
        *
        *  X X X 
        *  X
        *
        *  X
        *
        * This is because pixels on the top and left boundaries of polygons are drawn, but not on the bottom and 
        * right.  So, if you have a line going up that turns the corner and goes right, there is a one pixel shift 
        * in the pattern.
        *
        * So, to fix this, we drawn the top and right in one call, then the left and bottom in another call, fixing 
        * up the dash offset for the second call ourselves to get continuity at the upper left.
        *
        * It's not perfect since we really should have a join at the upper left and lower right instead of two
        * intersecting lines but that's only really apparent for no-dashes, which(for this reason) are done as one
        * polygon and don't to through this code path.
        */
        
        dash_len = strlen(dash_list);
        
        if(dash_list[0])
            gdk_gc_set_dashes(gc, 0, dash_list, dash_len);
        
        gdk_draw_lines(window, gc, points, 3);
        
        /* We draw this line one farther over than it is "supposed" to because of another rasterization problem ... 
        * if two 1 pixel unjoined lines meet at the lower right, there will be a missing pixel.
        */
        points[2].x += 1;
        
        if(dash_list[0])
        {
            gint dash_pixels = 0;
            gint i;
        
            /* Adjust the dash offset for the bottom and left so we match up at the upper left. */
            for(i = 0; i < dash_len; i++)
                dash_pixels += dash_list[i];
        
            if(dash_len % 2 == 1)
                dash_pixels *= 2;
        
            gdk_gc_set_dashes(gc, dash_pixels -(width + height - 2 * line_width) % dash_pixels, dash_list, dash_len);
        }
        
        gdk_draw_lines(window, gc, points + 2, 3);
    }
    
    gdk_gc_set_line_attributes(gc, 0, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
    
    if(area)
        gdk_gc_set_clip_rectangle(gc, NULL);
    
    if(free_dash_list)
        g_free(dash_list);
}

#if GTK_MAJOR_VERSION>1
static void draw_resize_grip(GtkStyle *style, GdkWindow *window, GtkStateType state, GdkRectangle *area, 
                             GtkWidget *widget, const gchar *detail, GdkWindowEdge edge, gint x, gint y, gint width,
                             gint height)
{
    QT_CURVE_STYLE_VAR
    FN_CHECK
  
    if(area)
    {
        gdk_gc_set_clip_rectangle(style->light_gc[state], area);
        gdk_gc_set_clip_rectangle(style->dark_gc[state], area);
        gdk_gc_set_clip_rectangle(style->bg_gc[state], area);
    }

    switch(edge)
    {
        case GDK_WINDOW_EDGE_NORTH_WEST:
            /* make it square */
            if(width < height)
                height = width;
            else if(height < width)
                width = height;
            break;
        case GDK_WINDOW_EDGE_NORTH:
            if(width < height)
                height = width;
            break;
        case GDK_WINDOW_EDGE_NORTH_EAST:
            /* make it square, aligning to top right */
            if(width < height)
                height = width;
            else if(height < width)
            {
                x +=(width - height);
                width = height;
            }
            break;
        case GDK_WINDOW_EDGE_WEST:
            if(height < width)
                width = height;
            break;
        case GDK_WINDOW_EDGE_EAST:
            /* aligning to right */
            if(height < width)
            {
                x +=(width - height);
                width = height;
            }
            break;
        case GDK_WINDOW_EDGE_SOUTH_WEST:
            /* make it square, aligning to bottom left */
            if(width < height)
            {
                y +=(height - width);
                height = width;
            }
            else if(height < width)
                width = height;
            break;
        case GDK_WINDOW_EDGE_SOUTH:
            /* align to bottom */
            if(width < height)
            {
                y +=(height - width);
                height = width;
            }
            break;
        case GDK_WINDOW_EDGE_SOUTH_EAST:
            /* make it square, aligning to bottom right */
            if(width < height)
            {
                y +=(height - width);
                height = width;
            }
            else if(height < width)
            {
                x +=(width - height);
                width = height;
            }
            break;
        default:
            g_assert_not_reached();
    }

  /* Clear background */
  gtk_style_apply_default_background(style, window, FALSE, state, area, x, y, width, height);   

    switch(edge)
    {
        case GDK_WINDOW_EDGE_WEST:
        case GDK_WINDOW_EDGE_EAST:
        {
            gint xi = x;

            while(xi < x + width)
            {
                gdk_draw_line(window, qtcurve_style->gray_gc[0], xi, y, xi, y + height);
                xi++;
                gdk_draw_line(window, qtcurve_style->gray_gc[5], xi, y, xi, y + height);
                xi += 2;
            }
            break;
        }
        case GDK_WINDOW_EDGE_NORTH:
        case GDK_WINDOW_EDGE_SOUTH:
        {
            gint yi = y;
    
            while(yi < y + height)
            {
                gdk_draw_line(window, qtcurve_style->gray_gc[0], x, yi, x + width, yi);
                yi++;
                gdk_draw_line(window, qtcurve_style->gray_gc[5], x, yi, x + width, yi);
                yi+= 2;
            }
            break;
        }
        case GDK_WINDOW_EDGE_NORTH_WEST:
        {
            gint xi = x + width,
                 yi = y + height;
        
            while(xi > x + 3)
            {
                gdk_draw_line(window, qtcurve_style->gray_gc[5], xi, y, x, yi);
                --xi;
                --yi;
                gdk_draw_line(window, qtcurve_style->gray_gc[0], xi, y, x, yi);
                xi -= 3;
                yi -= 3;
            }
            break;
        }
        case GDK_WINDOW_EDGE_NORTH_EAST:
        {
            gint xi = x,
                 yi = y + height;
    
            while(xi <(x + width - 3))
            {
                gdk_draw_line(window, qtcurve_style->gray_gc[0], xi, y, x + width, yi);                           
                ++xi;
                --yi;
                gdk_draw_line(window, qtcurve_style->gray_gc[5], xi, y, x + width, yi);
                xi += 3;
                yi -= 3;
            }
            break;
        }
        case GDK_WINDOW_EDGE_SOUTH_WEST:
        {    
            gint xi = x + width,
            yi = y;
    
            while(xi > x + 3)
            {
                gdk_draw_line(window, qtcurve_style->gray_gc[5], x, yi, xi, y + height);
                --xi;
                ++yi;
                gdk_draw_line(window, qtcurve_style->gray_gc[0], x, yi, xi, y + height);
                xi -= 3;
                yi += 3;
            }
            break;
        }
        case GDK_WINDOW_EDGE_SOUTH_EAST:
        {
            gint xi = x,
                 yi = y;
    
            while(xi <(x + width - 3))
            {
                gdk_draw_line(window, qtcurve_style->gray_gc[0], xi, y + height, x + width, yi);
                ++xi;
                ++yi;
                gdk_draw_line(window, qtcurve_style->gray_gc[5], xi, y + height, x + width, yi);
                xi += 3;
                yi += 3;
            }
            break;
        }
        default:
            g_assert_not_reached();
    }
  
    if(area)
    {
        gdk_gc_set_clip_rectangle(style->light_gc[state], NULL);
        gdk_gc_set_clip_rectangle(style->dark_gc[state], NULL);
        gdk_gc_set_clip_rectangle(style->bg_gc[state], NULL);
    }
}

static void draw_expander(GtkStyle *style, GdkWindow *window, GtkStateType state, GdkRectangle *area,
                          GtkWidget *widget, const gchar *detail, gint x, gint y, GtkExpanderStyle expander_style)
{
    GdkGC *gc=style->text_gc[QTC_ARROW_STATE(state)];
    int   lv_size=QTC_LV_SIZE(global.lv_expander);
#if 0
printf("Draw expander %d %s  ", state, detail ? detail : "NULL");
display_widget(widget, 5);
#endif

    x-=lv_size>>1;
    y-=lv_size>>1;

    if(LV_LINES_NONE!=global.lv_lines)
    {
        GdkGC *mid_gc=style->text_gc[ /* global.lv_dark ? GTK_STATE_NORMAL : */ GTK_STATE_INSENSITIVE];
        int   lo=global.rounded ? 2 : 0;

        gdk_draw_line(window, mid_gc, (x-2)+lo, y-2, (x+2+lv_size-1)-lo, y-2);
        gdk_draw_line(window, mid_gc, (x-2)+lo, y+2+lv_size-1, (x+2+lv_size-1)-lo, y+2+lv_size-1);
        gdk_draw_line(window, mid_gc, x-2, (y-2)+lo, x-2, (y+2+lv_size-1)-lo);
        gdk_draw_line(window, mid_gc, x+2+lv_size-1, (y-2)+lo, x+2+lv_size-1, (y+2+lv_size-1)-lo);

        if(global.rounded)
        {
            gdk_draw_point(window, mid_gc, x-1, y-1);
            gdk_draw_point(window, mid_gc, x-1, y+lv_size);
            gdk_draw_point(window, mid_gc, x+lv_size, y-1);
            gdk_draw_point(window, mid_gc, x+lv_size, y+lv_size);
/*
CPD: Can't tell when an item is selected - therefore AA looks bad, :-(
            QTC_SET_MID_COLOR(style->base_gc[state],
                              style->text[global.lv_dark ? GTK_STATE_NORMAL : GTK_STATE_INSENSITIVE], style->bg[state])
            gdk_draw_line(window, qtc_mid_color_gc, x-2, y-1, x-1, y-2);
            gdk_draw_line(window, qtc_mid_color_gc, x+lv_size, y-2, x+lv_size+1, y-1);
            gdk_draw_line(window, qtc_mid_color_gc, x-2, y+lv_size, x-1, y+lv_size+1);
            gdk_draw_line(window, qtc_mid_color_gc, x+lv_size, y+lv_size+1, x+lv_size+1, y+lv_size);
            QTC_UNSET_MID_COLOR
*/
        }
    }

    if(LV_EXP_ARR==global.lv_expander)
    {
        if(GTK_EXPANDER_COLLAPSED==expander_style)
            draw_arrow(window, gc, area,
                       GTK_TEXT_DIR_RTL==gtk_widget_get_direction(widget) ? GTK_ARROW_LEFT : GTK_ARROW_RIGHT,
                       x+2, y, LARGE_ARR_HEIGHT, LARGE_ARR_WIDTH);
        else
            draw_arrow(window, gc, area, GTK_ARROW_DOWN, x, y+2, LARGE_ARR_WIDTH, LARGE_ARR_HEIGHT);
    }
    else
    {
        int mid=lv_size>>1;

        gdk_draw_line(window, gc, x+(mid-2), y+mid, x+lv_size-(mid-1), y+mid);
        if(GTK_EXPANDER_COLLAPSED==expander_style)
            gdk_draw_line(window, gc, x+mid, y+(mid-2), x+mid, y+lv_size-(mid-1));
    }
}

#endif

static GdkGC * realize_color(GtkStyle *style, GdkColor *color)
{
    GdkGCValues gc_values;

    gdk_colormap_alloc_color(style->colormap, color, FALSE, TRUE);
    gc_values.foreground = *color;

    return gtk_gc_get(style->depth, style->colormap, &gc_values, GDK_GC_FOREGROUND);
}

#if GTK_MAJOR_VERSION>1
static void qtcurve_style_realize(GtkStyle *style)
#else
static void theme_realize_style(GtkStyle *style)
#endif
{
    QT_CURVE_STYLE_VAR
    int i;

#if GTK_MAJOR_VERSION>1
    parent_class->realize(style);
#endif

    GEN_GCS(style, qtcurve_style->gray, qtcurve_style->gray_gc);
    GEN_GCS(style, qtcurve_style->button, qtcurve_style->button_gc);

    for(i=0; i<GRADIENT_NUM_COLS-2; ++i)
        qtcurve_style->gradient_gc[i] = realize_color(style, &qtcurve_style->gradient[i]);
}

#if GTK_MAJOR_VERSION>1
static void qtcurve_style_unrealize(GtkStyle *style)
#else
static void theme_unrealize_style(GtkStyle *style)
#endif
{
    QT_CURVE_STYLE_VAR
    int i;
  
    /* We don't free the colors, because we don't know if gtk_gc_release() actually freed the GC. FIXME - need
     * a way of ref'ing colors explicitely so GtkGC can handle things properly.
     */
    RELEASE_GCS(qtcurve_style->gray_gc);
    RELEASE_GCS(qtcurve_style->button_gc);

    for(i=0; i<GRADIENT_NUM_COLS-2; ++i)
        gtk_gc_release(qtcurve_style->gradient_gc[i]);
}

#if GTK_MAJOR_VERSION>1
static void qtcurve_style_init_from_rc(GtkStyle *style, GtkRcStyle *rc_style)
{
    QT_CURVE_STYLE_VAR

    parent_class->init_from_rc(style, rc_style);

    shade_colors(&qt_settings.colors[COLOR_BUTTON], qtcurve_style->button);
    shade_colors(&qt_settings.colors[COLOR_WINDOW], qtcurve_style->gray);

    qtcurve_style->gradient[GRADIENT_BASE]=qt_settings.colors[COLOR_SELECTED];
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_TOP], SHADE_GRADIENT_TOP);
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_BOTTOM], SHADE_GRADIENT_BOTTOM);
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_LIGHT], SHADE_GRADIENT_LIGHT);
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_DARK], SHADE_GRADIENT_DARK);
}
#endif

#if GTK_MAJOR_VERSION>1
void qtcurve_style_class_init(QtCurveStyleClass *klass)
#else
void qtcurve_initialize_style(GtkStyleClass *style_class, GtkStyleClass *parent)
#endif
{
#if GTK_MAJOR_VERSION>1
    GtkStyleClass *style_class = GTK_STYLE_CLASS(klass);

    parent_class = g_type_class_peek_parent(klass);
#else
    parent_class = parent;    /* Save for later use */
    *style_class = *parent;
#endif

#if GTK_MAJOR_VERSION>1
    style_class->realize = qtcurve_style_realize;
    style_class->unrealize = qtcurve_style_unrealize;
    style_class->init_from_rc = qtcurve_style_init_from_rc;
    style_class->draw_resize_grip = draw_resize_grip;
    style_class->draw_expander = draw_expander;
#endif
    style_class->draw_arrow = qtcurve_draw_arrow;
    style_class->draw_tab = draw_tab;
    style_class->draw_shadow = draw_shadow;
    style_class->draw_box_gap = draw_box_gap;
    style_class->draw_extension = draw_extension;
    style_class->draw_handle = draw_handle;
    style_class->draw_box = draw_box;
    style_class->draw_flat_box = draw_flat_box;
    style_class->draw_check = draw_check;
    style_class->draw_slider = draw_slider;
    style_class->draw_option = draw_option;
    style_class->draw_shadow_gap = draw_shadow_gap;
    style_class->draw_hline = draw_hline;
    style_class->draw_vline = draw_vline;
    style_class->draw_focus = draw_focus;
#if GTK_MAJOR_VERSION>1
    style_class->draw_layout = draw_layout;
#else
    style_class->draw_string = draw_string;
#endif
}

#if GTK_MAJOR_VERSION>1
static GtkRcStyleClass *parent_rc_class;
GType qtcurve_type_rc_style = 0;
#endif

#if GTK_MAJOR_VERSION==1
static void qtcurve_rc_data_ref(QtCurveRcStyle *qtcurve_style)
{
    qtcurve_style->refcount++;
}

static void qtcurve_rc_data_unref(QtCurveRcStyle *qtcurve_style)
{
    qtcurve_style->refcount--;
    if(0==qtcurve_style->refcount)
        g_free(qtcurve_style);
}
#endif

enum
{
    TOKEN_THICKNESS_HACK = G_TOKEN_LAST + 1
#ifndef USE_SINGLE_STYLE
  , TOKEN_BORDER_FRAME,
    TOKEN_BORDER_BUTTON,
    TOKEN_TOOLBAR_BORDERS,
    TOKEN_ROUNDED,
    TOKEN_SLIDER_THUMBS,
    TOKEN_HANDLES,
    TOKEN_GRADIENT_EFFECT,
    TOKEN_GRADIENT_EFFECT_PM,
    TOKEN_V_ARROW,
    TOKEN_BOLD_DEF_TEXT,
    TOKEN_DEF_BTN_INDICATOR,
    TOKEN_RADIO_CHECK_LABEL_HIGHLIGHT,
    TOKEN_LV_EXPANDER,
    TOKEN_LV_LINES,
/*    TOKEN_LV_DARK, */
    TOKEN_TRUE,
    TOKEN_FALSE
#endif
#if GTK_MAJOR_VERSION==1
  , TOKEN_XTHICK,
    TOKEN_YTHICK
#endif
};

static struct
{
    const gchar *name;
    guint       token;
}
theme_symbols[] =
{
#ifndef USE_SINGLE_STYLE
    { "border_frame",                TOKEN_BORDER_FRAME },
    { "border_button",               TOKEN_BORDER_BUTTON },
    { "toolbar_borders",             TOKEN_TOOLBAR_BORDERS },
    { "rounded",                     TOKEN_ROUNDED },
    { "slider_thumbs",               TOKEN_SLIDER_THUMBS },
    { "handles",                     TOKEN_HANDLES },
    { "gradient_effect",             TOKEN_GRADIENT_EFFECT },
    { "gradient_effect_pm",          TOKEN_GRADIENT_EFFECT_PM },
    { "v_arrow",                     TOKEN_V_ARROW },
    { "bold_def_text",               TOKEN_BOLD_DEF_TEXT },
    { "def_btn_indicator",           TOKEN_DEF_BTN_INDICATOR },
    { "radio_check_label_highlight", TOKEN_RADIO_CHECK_LABEL_HIGHLIGHT },
#if GTK_MAJOR_VERSION>1
    { "lv_expander",                 TOKEN_LV_EXPANDER },
    { "lv_lines",                    TOKEN_LV_LINES },
    /*{ "lv_dark",                     TOKEN_LV_DARK },*/
#endif
    { "TRUE",                        TOKEN_TRUE },
    { "FALSE",                       TOKEN_FALSE },
#endif
#if GTK_MAJOR_VERSION==1
    { "xthickness",                  TOKEN_XTHICK  },
    { "ythickness",                  TOKEN_YTHICK  },
#endif
    { "thickness_from_frame_border", TOKEN_THICKNESS_HACK }
};

static guint n_theme_symbols = sizeof(theme_symbols) / sizeof(theme_symbols[0]);

#if GTK_MAJOR_VERSION==1
static guint qtcurve_theme_parse_int(GScanner *scanner, int *i)
{
    guint token = g_scanner_get_next_token(scanner);
    token = g_scanner_get_next_token(scanner);

    if(token != G_TOKEN_EQUAL_SIGN)
        return G_TOKEN_EQUAL_SIGN;
    
    token = g_scanner_get_next_token(scanner);
    
    if(token != G_TOKEN_INT && token != G_TOKEN_STRING)
        return G_TOKEN_STRING;
    *i = G_TOKEN_STRING==token ? atoi(scanner->value.v_string) : scanner->value.v_int;
    
    return G_TOKEN_NONE;
}
#endif

#ifndef USE_SINGLE_STYLE
static guint qtcurve_theme_parse_bool(GScanner *scanner, gboolean *b)
{
  guint token = g_scanner_get_next_token(scanner);
  token = g_scanner_get_next_token(scanner);

  if(token != G_TOKEN_EQUAL_SIGN)
      return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);

  if(token == TOKEN_TRUE)
      *b = TRUE;
  else if(token == TOKEN_FALSE)
      *b = FALSE;
  else
      return TOKEN_TRUE;

  return G_TOKEN_NONE;
}

static guint qtcurve_theme_parse_def_btn_indicator(GScanner *scanner, EDefBtnIndicator *dbi)
{
  guint token = g_scanner_get_next_token(scanner);

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_STRING)
    return G_TOKEN_STRING;

  *dbi=qtc_to_ind(scanner->value.v_string);

  return G_TOKEN_NONE;
}

static guint qtcurve_theme_parse_groove(GScanner *scanner, EGroove *groove)
{
  guint token = g_scanner_get_next_token(scanner);

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_STRING)
    return G_TOKEN_STRING;

  *groove=qtc_to_groove(scanner->value.v_string);

  return G_TOKEN_NONE;
}

static guint qtcurve_theme_parse_tbar_border(GScanner *scanner, ETBarBorder *border)
{
  guint token = g_scanner_get_next_token(scanner);

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_STRING)
    return G_TOKEN_STRING;

  *border=qtc_to_tbar_border(scanner->value.v_string);

  return G_TOKEN_NONE;
}

#if GTK_MAJOR_VERSION>1
static guint qtcurve_theme_parse_lv_expander(GScanner *scanner, ELvExpander *exp)
{
  guint token = g_scanner_get_next_token(scanner);

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_STRING)
    return G_TOKEN_STRING;

  *exp=qtc_to_tbar_lv_expander(scanner->value.v_string);

  return G_TOKEN_NONE;
}

static guint qtcurve_theme_parse_lv_lines(GScanner *scanner, ELvLines *lines)
{
  guint token = g_scanner_get_next_token(scanner);

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_EQUAL_SIGN)
    return G_TOKEN_EQUAL_SIGN;

  token = g_scanner_get_next_token(scanner);
  if (token != G_TOKEN_STRING)
    return G_TOKEN_STRING;

  *lines=qtc_to_tbar_lv_lines(scanner->value.v_string);

  return G_TOKEN_NONE;
}
#endif

#endif

#if GTK_MAJOR_VERSION>1
static guint qtcurve_rc_style_parse(GtkRcStyle *rc_style, GtkSettings *settings, GScanner *scanner)
#else
static guint theme_parse_rc_style(GScanner *scanner, GtkRcStyle *rc_style)
#endif
{
    static GQuark scope_id = 0;
#if GTK_MAJOR_VERSION>1
    QtCurveRcStyle *qtcurve_style = QTCURVE_RC_STYLE(rc_style);
#else
    QtCurveRcStyle *qtcurve_style;
#endif
    guint old_scope,
          token;
  
    /* Set up a new scope in this scanner. */
    if(!scope_id)
        scope_id = g_quark_from_string("qtcurve_theme_engine");

    /* If we bail out due to errors, we *don't* reset the scope, so the error messaging code can make sense of our 
       tokens. */
    old_scope = g_scanner_set_scope(scanner, scope_id);

  /* We're ready to go, now parse the top level */

    if(!g_scanner_lookup_symbol(scanner, theme_symbols[0].name))
    {
        int i;

        g_scanner_freeze_symbol_table(scanner);
        for(i = 0; i < n_theme_symbols; i++)
            g_scanner_scope_add_symbol(scanner, scope_id, theme_symbols[i].name, 
                                       GINT_TO_POINTER(theme_symbols[i].token));
        g_scanner_thaw_symbol_table(scanner);
    }

#if GTK_MAJOR_VERSION==1
    qtcurve_style = g_new(QtCurveRcStyle, 1);
    qtcurve_style->refcount = 1;
    qtcurve_style->xthickness=qtcurve_style->ythickness=1;
#endif
    qt_set_font(rc_style);

    token = g_scanner_peek_next_token(scanner);
    while(token != G_TOKEN_RIGHT_CURLY)
    {
        switch(token)
        {
#ifndef USE_SINGLE_STYLE
            case TOKEN_BORDER_FRAME:
                token = qtcurve_theme_parse_bool(scanner, &(global.border_frame));
                break;
            case TOKEN_BORDER_BUTTON:
                token = qtcurve_theme_parse_bool(scanner, &(global.border_button));
                break;
            case TOKEN_TOOLBAR_BORDERS:
                token = qtcurve_theme_parse_tbar_border(scanner, &(global.toolbar_borders));
                break;
            case TOKEN_RADIO_CHECK_LABEL_HIGHLIGHT:
                token = qtcurve_theme_parse_bool(scanner, &(global.radio_check_label_highlight));
                break;
            case TOKEN_ROUNDED:
                token = qtcurve_theme_parse_bool(scanner, &(global.rounded));
                break;
            case TOKEN_SLIDER_THUMBS:
                token = qtcurve_theme_parse_groove(scanner, &(global.slider_thumbs));
                break;
            case TOKEN_HANDLES:
                token = qtcurve_theme_parse_groove(scanner, &(global.handles));
                break;
            case TOKEN_GRADIENT_EFFECT:
                token = qtcurve_theme_parse_bool(scanner, &(global.gradient_effect));
                break;   
            case TOKEN_GRADIENT_EFFECT_PM:
                token = qtcurve_theme_parse_bool(scanner, &(global.gradient_effect_pm));
                break;
            case TOKEN_V_ARROW:
                token = qtcurve_theme_parse_bool(scanner, &(global.v_arrow));
                break;
            case TOKEN_BOLD_DEF_TEXT:
                token = qtcurve_theme_parse_bool(scanner, &(global.bold_def_text));
                break;
            case TOKEN_DEF_BTN_INDICATOR:
                token = qtcurve_theme_parse_def_btn_indicator(scanner, &(global.def_btn_indicator));
                break;
#if GTK_MAJOR_VERSION>1
            case TOKEN_LV_EXPANDER:
                token = qtcurve_theme_parse_lv_expander(scanner, &(global.lv_expander));
                break;
            case TOKEN_LV_LINES:
                token = qtcurve_theme_parse_lv_lines(scanner, &(global.lv_lines));
                break;
/*
            case TOKEN_LV_DARK:
                token = qtcurve_theme_parse_bool(scanner, &(global.lv_dark));
                break;
*/
#endif
#endif
#if GTK_MAJOR_VERSION==1
            case TOKEN_XTHICK:
                token = qtcurve_theme_parse_int(scanner, &(qtcurve_style->xthickness));
                break;
            case TOKEN_YTHICK:
                token = qtcurve_theme_parse_int(scanner, &(qtcurve_style->ythickness));
                break;
#endif
           case TOKEN_THICKNESS_HACK:
                g_scanner_get_next_token(scanner);  /* Ignore... */
                token=G_TOKEN_NONE;
#if GTK_MAJOR_VERSION>1
                rc_style->xthickness=rc_style->ythickness=global.border_frame ? 2 : 1;
#else
                qtcurve_style->xthickness=qtcurve_style->ythickness=global.border_frame ? 2 : 1;
#endif
                break;
           default:
               g_scanner_get_next_token(scanner); 
               token = G_TOKEN_RIGHT_CURLY;
        }

        if(token != G_TOKEN_NONE)
            return token;

        token = g_scanner_peek_next_token(scanner);
    }

    g_scanner_get_next_token(scanner);

#if GTK_MAJOR_VERSION==1
    rc_style->engine_data = qtcurve_style;
#endif
    g_scanner_set_scope(scanner, old_scope);

    return G_TOKEN_NONE;
}

#if GTK_MAJOR_VERSION>1
static void qtcurve_rc_style_merge(GtkRcStyle *dest, GtkRcStyle *src)
{
    QtCurveRcStyle *dest_w,
                   *src_w;
  
    parent_rc_class->merge(dest, src);

    if(!QTCURVE_IS_RC_STYLE(src))
        return;
  
    src_w = QTCURVE_RC_STYLE(src);
    dest_w = QTCURVE_RC_STYLE(dest);
}
#else
static void theme_merge_rc_style(GtkRcStyle *dest, GtkRcStyle *src)
{
    QtCurveRcStyle *src_data = src->engine_data;
  
    if(!dest->engine_data && src_data)
    {
        qtcurve_rc_data_ref(src_data);
        dest->engine_data = src_data;
    }
}
#endif

#if GTK_MAJOR_VERSION==1
static void set_props(GtkStyle *style)
{
  /* TODO: Change and/or add to these: */
#define OPTION_INDICATOR_WIDTH 7
#define OPTION_INDICATOR_LEFT_SPACING 7
#define OPTION_INDICATOR_RIGHT_SPACING 5
    gtk_style_set_prop_experimental(style, "GtkButton::default_spacing", 6);
    gtk_style_set_prop_experimental(style, "GtkCheckButton::indicator_size", QTC_CHECK_SIZE);
    gtk_style_set_prop_experimental(style, "GtkOptionMenu::indicator_width", OPTION_INDICATOR_WIDTH);
    gtk_style_set_prop_experimental(style, "GtkOptionMenu::indicator_left_spacing", OPTION_INDICATOR_LEFT_SPACING);
    gtk_style_set_prop_experimental(style, "GtkOptionMenu::indicator_right_spacing", OPTION_INDICATOR_RIGHT_SPACING);
    gtk_style_set_prop_experimental(style, "GtkPaned::handle_full_size", 1);
    gtk_style_set_prop_experimental(style, "GtkPaned::handle_width", 6);
    gtk_style_set_prop_experimental(style, "GtkRange::always_draw_trough", 1);
    gtk_style_set_prop_experimental(style, "GtkRange::trough_border", 0);
    gtk_style_set_prop_experimental(style, "GtkRange::slider_width", 15);
    gtk_style_set_prop_experimental(style, "GtkRange::stepper_size", 15);
    gtk_style_set_prop_experimental(style, "GtkRange::stepper_spacing", 0);
    gtk_style_set_prop_experimental(style, "GtkRange::always_redraw", 1);
    gtk_style_set_prop_experimental(style, "GtkScale::slider_width", 15);
    gtk_style_set_prop_experimental(style, "GtkScrollbar::min_slider_length", 16);
    gtk_style_set_prop_experimental(style, "GtkSpinButton::shadow_type", GTK_SHADOW_IN);
}

static void theme_rc_style_to_style(GtkStyle *style, GtkRcStyle *rc_style)
{
    static GtkStyleClass *default_class = NULL;
    GtkStyleClass        *class=NULL;
    QtCurveStyle         *qtcurve_style;
    QtCurveRcStyle       *qtcurve_rc_style=rc_style->engine_data;
    
    if(!default_class)
    {
        default_class = g_new(GtkStyleClass, 1);
        qtcurve_initialize_style(default_class, style->klass);
        default_class->xthickness = 1;
        default_class->ythickness = 1;
    }
    
    if(qtcurve_rc_style->xthickness!=default_class->xthickness || 
       qtcurve_rc_style->ythickness!=default_class->ythickness)
    {
        class = g_new(GtkStyleClass, 1);
        qtcurve_initialize_style(class, style->klass);
        class->xthickness = qtcurve_rc_style->xthickness;
        class->ythickness = qtcurve_rc_style->ythickness;
    }
    else
        class=default_class;
    
    style->klass = class;
    qtcurve_style = style->engine_data = g_new0(QtCurveStyle, 1);
    
    shade_colors(&qt_settings.colors[COLOR_WINDOW], qtcurve_style->gray);
    shade_colors(&qt_settings.colors[COLOR_BUTTON], qtcurve_style->button);
    
    qtcurve_style->gradient[GRADIENT_BASE]=qt_settings.colors[COLOR_SELECTED];
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_TOP], SHADE_GRADIENT_TOP);
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_BOTTOM], SHADE_GRADIENT_BOTTOM);
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_LIGHT], SHADE_GRADIENT_LIGHT);
    shade(&qt_settings.colors[COLOR_SELECTED], &qtcurve_style->gradient[GRADIENT_DARK], SHADE_GRADIENT_DARK);
    
    set_props(style);
#ifdef QTC_HIGHLIGHT_CHECK_RADIO_TEXT
    qt_set_colours(style, rc_style, 12);
#else
    qt_set_colours(style, rc_style);
#endif
}

static void theme_duplicate_style(GtkStyle *dest, GtkStyle *src)
{
    dest->engine_data = g_memdup(src->engine_data, sizeof(QtCurveStyle));
    set_props(dest);
}
#else
/* Create an empty style suitable to this RC style */
static GtkStyle * qtcurve_rc_style_create_style(GtkRcStyle *rc_style)
{
    GtkStyle *style=g_object_new(QTCURVE_TYPE_STYLE, NULL);

#ifdef QTC_HIGHLIGHT_CHECK_RADIO_TEXT
    qt_set_colours(style, rc_style, 12);
#else
    qt_set_colours(style, rc_style);
#endif

    return style;
}
#endif

#if GTK_MAJOR_VERSION==1
static void theme_destroy_rc_style(GtkRcStyle *rc_style)
{
    QtCurveRcStyle *rc_data = rc_style->engine_data;
    qtcurve_rc_data_unref(rc_data);
}

static void theme_destroy_style(GtkStyle *style)
{
    g_free(style->engine_data);
}
#endif

#if GTK_MAJOR_VERSION>1
GType qtcurve_type_style = 0;

static void qtcurve_style_init(QtCurveStyle *style)
{
}

void qtcurve_style_register_type(GTypeModule *module)
{
    static const GTypeInfo object_info =
    {
        sizeof(QtCurveStyleClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) qtcurve_style_class_init,
        NULL,            /* class_finalize */
        NULL,            /* class_data */
        sizeof(QtCurveStyle),
        0,                /* n_preallocs */
        (GInstanceInitFunc) qtcurve_style_init,
         NULL
    };

    qtcurve_type_style = g_type_module_register_type(module, GTK_TYPE_STYLE, "QtCurveStyle", &object_info, 0);
}

static void qtcurve_rc_style_init(QtCurveRcStyle *qtcurve_rc)
{
    qt_init(&global);
}

static void qtcurve_rc_style_finalize(GObject *object)
{
    qt_exit();
}

static void qtcurve_rc_style_class_init(QtCurveRcStyleClass *klass)
{
    GtkRcStyleClass *rc_style_class = GTK_RC_STYLE_CLASS(klass);
    GObjectClass    *object_class = G_OBJECT_CLASS(klass);
    
    parent_rc_class = g_type_class_peek_parent(klass);
    
    rc_style_class->parse = qtcurve_rc_style_parse;
    rc_style_class->create_style = qtcurve_rc_style_create_style;
    rc_style_class->merge = qtcurve_rc_style_merge;
    
    object_class->finalize = qtcurve_rc_style_finalize;
}

void qtcurve_rc_style_register_type(GTypeModule *module)
{
    static const GTypeInfo object_info =
    {
        sizeof(QtCurveRcStyleClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) qtcurve_rc_style_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof(QtCurveRcStyle),
        0,              /* n_preallocs */
        (GInstanceInitFunc) qtcurve_rc_style_init,
        NULL
    };
  
    qtcurve_type_rc_style = g_type_module_register_type(module, GTK_TYPE_RC_STYLE, "QtCurveRcStyle", &object_info, 0);
}
#endif

#if GTK_MAJOR_VERSION>1
G_MODULE_EXPORT void theme_init(GTypeModule *module)
#else
G_MODULE_EXPORT void theme_init(GtkThemeEngine *engine)
#endif
{
    global.contrast=7;
#ifdef USE_SINGLE_STYLE
    /* V3 is default */
    global.rounded=global.border_frame=global.border_button=global.gradient_effect=
    global.gradient_effect_pm=global.bold_def_text=TRUE;
    global.v_arrow=global.radio_check_label_highlight;
    global.def_btn_indicator=IND_CORNER;
    global.slider_thumbs=global.handles=GROOVE_SUNKEN;
    global.toolbar_borders=TB_LIGHT;
#if GTK_MAJOR_VERSION>1
    global.lv_expander=LV_EXP_ARR;
    global.lv_lines=LV_LINES_SOLID;
/*    global.lv_dark=FALSE; */
#endif
#else
    /* V1 is default */
    global.rounded=global.border_frame=global.border_button=global.gradient_effect=
    global.gradient_effect_pm=global.bold_def_text=FALSE;
    global.v_arrow=global.radio_check_label_highlight=TRUE;
    global.def_btn_indicator=IND_BORDER;
    global.slider_thumbs=global.handles=GROOVE_RAISED;
    global.toolbar_borders=TB_DARK;
#if GTK_MAJOR_VERSION>1
    global.lv_expander=LV_EXP_PM;
    global.lv_lines=LV_LINES_DOTTED;
/*    global.lv_dark=TRUE; */
#endif
#endif
#if GTK_MAJOR_VERSION>1
    qtcurve_rc_style_register_type(module);
    qtcurve_style_register_type(module);
#else
    engine->parse_rc_style = theme_parse_rc_style;
    engine->merge_rc_style = theme_merge_rc_style;
    engine->rc_style_to_style = theme_rc_style_to_style;
    engine->duplicate_style = theme_duplicate_style;
    engine->realize_style = theme_realize_style;
    engine->unrealize_style = theme_unrealize_style;
    engine->destroy_rc_style = theme_destroy_rc_style;
    engine->destroy_style = theme_destroy_style;
    engine->set_background = NULL;
    
    gdk_rgb_init();
    qt_init(&global);
#endif
}

G_MODULE_EXPORT void theme_exit()
{
#if GTK_MAJOR_VERSION==1
    qt_exit();
#endif
}

#if GTK_MAJOR_VERSION>1
G_MODULE_EXPORT GtkRcStyle * theme_create_rc_style()
{
    return GTK_RC_STYLE(g_object_new(QTCURVE_TYPE_RC_STYLE, NULL));
}
#endif

/* The following function will be called by GTK+ when the module is loaded and checks to see if we are compatible with the
 * version of GTK+ that loads us. */
G_MODULE_EXPORT const gchar * g_module_check_init(GModule *module);

const gchar* g_module_check_init(GModule *module)
{
    return gtk_check_version(GTK_MAJOR_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION - GTK_INTERFACE_AGE);
}
